# -*- coding: utf-8 -*-
#
"""Testing script for wlsqm, doubles as a usage example.

-JJ 2016-11-10
"""

from __future__ import division
from __future__ import absolute_import

import time

import numpy as np
import sympy as sy

import scipy.spatial  # cKDTree

import pylab as pl
import mpl_toolkits.mplot3d.axes3d as p3

try:
    import wlsqm
except ImportError:
    print "WLSQM not found; is it installed?"
    from sys import exit
    exit(1)

import sudoku_lhs


# from various scripts, e.g. miniprojects/misc/tworods/main2.py
def axis_marginize(ax, epsx, epsy):
    a = ax.axis()
    w = a[1] - a[0]
    h = a[3] - a[2]
    ax.axis( [ a[0] - w*epsx, a[1] + w*epsx, a[2] - h*epsy, a[3] + h*epsy] )


# from find_neighbors2.py
class SimpleTimer:
    def __init__(self, label="", n=None):
        self.label = label
        self.n     = n      # number of repetitions done inside the "with..." section (for averaging in timing info)

    def __enter__(self):
        self.t0 = time.time()
        return self

    def __exit__(self, errtype, errvalue, traceback):
        dt         = time.time() - self.t0
        identifier = ("%s" % self.label) if len(self.label) else "time taken: "
        avg        = (", avg. %gs per run" % (dt/self.n)) if self.n is not None else ""
        print "%s%gs%s" % (identifier, dt, avg)


# many simultaneous local models, 2D
#
def testmany2d():
    #########################
    # config
    #########################

    ntasks = 8  # OpenMP parallelization

    axislims = [0., 1., 0., 1.]  # [xmin, xmax, ymin, ymax], for plotting
    nvis = 201  # number of visualization points per axis

    expr    = sy.sympify("sin(pi*x) * cos(pi*y)")

    points_per_axis = 100  # for point cloud generation

    r      = 5e-2  # neighborhood radius
    max_nk = 100   # maximum number of neighbor points to accept into each neighborhood (affects memory allocation)

    knowns = 1  # function value is known

    fit_order = 4
    weighting_method = wlsqm.WEIGHT_CENTER

    max_iter = 10  # for iterative fitting method

    reps = 20  # for demonstration of solving multiple times using the same geometry

    #########################
    # the test itself
    #########################

    print
    print "=" * 79
    print "many neighborhoods, 2D case"
    print "=" * 79
    print

    # create a stratified point cloud
    print "generating sudoku sample"
    with SimpleTimer(label=("    done in ")) as s:
        S,m = sudoku_lhs.sample(2, points_per_axis, 1)
        bins_per_axis = m*points_per_axis
        S = S / float(bins_per_axis - 1)  # scale the sample from [0, bins_per_axis-1]**2 to [0, 1]**2
        npoints = len(S)
        print "    %d points" % (npoints)

    # index the point cloud for fast neighbor searching
    print "indexing sample"
    with SimpleTimer(label=("    done in ")) as s:
        tree = scipy.spatial.cKDTree( data=S )

    # If this was an IBVP, we would here get the previous state of the unknown field.
    #
    # In this example, we just sample our function f().
    #
    lambdify_numpy_2d = lambda expr: sy.lambdify(("x","y"), expr, modules="numpy")  # SymPy expr --> lambda(x,y)
    f                 = lambdify_numpy_2d(expr)
    dfdx              = lambdify_numpy_2d(sy.diff(expr, "x"))
    dfdy              = lambdify_numpy_2d(sy.diff(expr, "y"))

    print "evaluating example function"
    with SimpleTimer(label=("    done in ")) as s:
        no = wlsqm.number_of_dofs( dimension=2, order=fit_order )
        fi = np.empty( (npoints,no), dtype=np.float64 )
        fi[:,0] = f( S[:,0], S[:,1] )  # fi[i,0] contains the function value at point S[i,:]

    # find the neighborhoods
    print "generating neighborhoods for each point"
    with SimpleTimer(label=("    done in ")) as s:
        hoods = np.zeros( (npoints,max_nk), dtype=np.int32 )  # neighbor point indices (pointing to rows in S[])
        nk    = np.empty( (npoints,), dtype=np.int32 )        # number of neighbors, i.e. nk[i] is the number of actually used columns in hoods[i,:]
        for i in xrange(npoints):
            I = tree.query_ball_point( S[i], r )  # indices of neighbors of S[i] at distance <= r  (but also including S[i] itself!)
            I = filter( lambda idx: idx != i, I )  # exclude S[i] itself
            if len(I) > max_nk:
                I = I[:max_nk]
            I = np.array( I, dtype=np.int32 )
            nk[i] = len(I)
            hoods[i,:nk[i]] = I

    # DEBUG
    print "number of neighbors min = %g, avg = %g, max = %g" % ( np.min(nk), np.mean(nk), np.max(nk) )
    print "neighbor lists for each problem instance:"
    print hoods
    print "number of neighbors for each problem instance:"
    print nk

    # perform the fitting
    print "fitting %d local surrogate models of order %d, driver mode (fit each model once)" % (npoints, fit_order)
    fit_order_array = fit_order        * np.ones( (npoints,), dtype=np.int32 )
    knowns_array    = knowns           * np.ones( (npoints,), dtype=np.int64 )
    wm_array        = weighting_method * np.ones( (npoints,), dtype=np.int32 )
    with SimpleTimer(label=("    done in ")) as s:
#        max_iterations_taken = wlsqm.fit_2D_many( xk=S[hoods], fk=fi[hoods,0], nk=nk,
#                                                   xi=S, fi=fi,
#                                                   sens=None, do_sens=False,
#                                                   order=fit_order_array, knowns=knowns_array, weighting_method=wm_array,
#                                                   debug=False )
        max_iterations_taken = wlsqm.fit_2D_many_parallel( xk=S[hoods], fk=fi[hoods,0], nk=nk,
                                                            xi=S, fi=fi,
                                                            sens=None, do_sens=False,
                                                            order=fit_order_array, knowns=knowns_array, weighting_method=wm_array,
                                                            ntasks=ntasks, debug=False )
#        max_iterations_taken = wlsqm.fit_2D_iterative_many( xk=S[hoods], fk=fi[hoods,0], nk=nk,
#                                                             xi=S, fi=fi,
#                                                             sens=None, do_sens=False,
#                                                             order=fit_order_array, knowns=knowns_array, weighting_method=wm_array,
#                                                             max_iter=max_iter, debug=False )
#        max_iterations_taken = wlsqm.fit_2D_iterative_many_parallel( xk=S[hoods], fk=fi[hoods,0], nk=nk,
#                                                            xi=S, fi=fi,
#                                                            sens=None, do_sens=False,
#                                                            order=fit_order_array, knowns=knowns_array, weighting_method=wm_array,
#                                                            max_iter=max_iter, ntasks=ntasks, debug=False )

    # Expert mode: allows solving multiple times (with new fk data) in the same geometry, performing the prepare step only once.
    #
    # This is especially good for a large number of repetitions with ALGO_BASIC, where a large majority of the computational cost comes from the prepare step.
    #
    # The total advantage is slightly smaller for a small number of repetitions with ALGO_ITERATIVE,
    # since the iterative mode already uses this strategy internally (also when invoked in driver mode).
    #
    print "fitting %d local surrogate models of order %d, expert mode" % (npoints, fit_order)
    print "    init"
    with SimpleTimer(label=("        done in ")) as s:
        solver = wlsqm.ExpertSolver( dimension=2, nk=nk, order=fit_order_array, knowns=knowns_array, weighting_method=wm_array, algorithm=wlsqm.ALGO_BASIC, do_sens=False, max_iter=max_iter, ntasks=ntasks, debug=False )
    print "    prepare"
    with SimpleTimer(label=("        done in ")) as s:
        solver.prepare( xi=S, xk=S[hoods] )
    print "    fit (each model %d times)" % (reps)
    with SimpleTimer(label=("        %d reps done in " % reps), n=reps) as s:
        for k in xrange(reps):
            solver.solve( fk=fi[hoods,0], fi=fi, sens=None )

    # DEBUG
    print "max corrective iterations taken: %d" % (max_iterations_taken)
    # see that we got the derivatives at each point
    if fit_order > 0:  # no derivatives if piecewise constant fit
        print dfdx( S[:,0], S[:,1] ) - fi[:,1]
        print dfdy( S[:,0], S[:,1] ) - fi[:,2]

    #########################
    # plotting
    #########################

    xx   = np.linspace(axislims[0], axislims[1], nvis)
    yy   = np.linspace(axislims[2], axislims[3], nvis)
    X,Y  = np.meshgrid(xx, yy)
    W    = f(X,Y)

    shp = np.shape(X)
    Xlin = np.reshape(X, -1)
    Ylin = np.reshape(Y, -1)
    x = np.empty( (len(Xlin), 2), dtype=np.float64 )
    x[:,0] = Xlin
    x[:,1] = Ylin
    print "preparing to interpolate global model"
    with SimpleTimer(label=("    done in ")) as s:
        solver.prep_interpolate()
    print "interpolating global model to %d points" % (len(Xlin))
    with SimpleTimer(label=("    done in ")) as s:
        W2,dummy = solver.interpolate( x, mode='continuous', r=r )  # slow, continuous
#        W2,dummy = solver.interpolate( x, mode='nearest' )  # fast, surprisingly accurate if a reasonable number of points (and continuous-looking although technically has jumps over Voronoi cell boundaries)
    W2 = np.reshape( W2, shp )

    # make 3d plot of the function
    #
    # see http://matplotlib.sourceforge.net/examples/mplot3d/lines3d_demo.html

    fig = pl.figure(3, figsize=(12,12))
    pl.clf()

    # Axes3D has a tendency to underestimate how much space it needs; it draws its labels
    # outside the window area in certain orientations.
    #
    # This causes the labels to be clipped, which looks bad. We prevent this by creating the axes
    # in a slightly smaller rect (leaving a margin). This way the labels will show - outside the Axes3D,
    # but still inside the figure window.
    #
    # The final touch is to set the window background to a matching white, so that the
    # background of the figure appears uniform.
    #
    fig.patch.set_color( (1,1,1) )
    fig.patch.set_alpha( 1.0 )
    x0y0wh = [ 0.02, 0.02, 0.96, 0.96 ]  # left, bottom, width, height      (here as fraction of subplot area)

#    # compute the corresponding figure coordinates for the 2x1 subplot layout
#    x0y0wh[0] = 0.5 + 0.5*x0y0wh[0]  # left
#    x0y0wh[2] = 0.5*x0y0wh[2]        # width

    ax = p3.Axes3D(fig, rect=x0y0wh)

    stride = max(1, (nvis-1)//10)  # pick a good-looking stride (for lines; we actually have more vertices, making a smoother-looking curve between the lines)
    # use linewidth=0 to remove the wireframe if desired.
#    surf = ax.plot_surface(X,Y,W, rstride=stride, cstride=stride, cmap=matplotlib.cm.Blues_r, clim=[fmin,fmax], linewidth=0.25, alpha=0.5)
    ax.plot_wireframe(X,Y,W, rstride=stride, cstride=stride, color='k', linewidth=0.5, linestyle='solid')
#        pl.colorbar(surf, shrink=0.5, aspect=5)
#        pl.colorbar(surf, shrink=0.96)

    # sampled points
    if points_per_axis < 50:
        ax.plot( S[:,0], S[:,1], f( S[:,0], S[:,1] ), linestyle='none', marker='o', markeredgecolor='r', markerfacecolor='none' )  # exact

    # surrogate model (global, patched)
    ax.plot_wireframe(X,Y,W2, rstride=stride, cstride=stride, color='r', linewidth=0.5, linestyle='solid')

#    ax.view_init(20, -48)
#    ax.view_init(18, -46)
#    ax.view_init(18, -128)
    ax.view_init(34, 140)

    ax.axis('tight')
    ax.set_zlim(-1.01, 1.01)
    pl.xlabel('$x$')
    pl.ylabel('$y$')
    ax.set_title('f(x,y)')


    print "    uninit"
    with SimpleTimer(label=("        done in ")) as s:
        del solver


# one local model, 3D
#
def test3d():
    #########################
    # config
    #########################

    axislims = [0., 1., 0., 1.]  # [xmin, xmax, ymin, ymax], for plotting
    nvis = 101  # number of visualization points per axis

    # Let's manufacture a solution (for which we know the derivatives analytically):
    #
    expr    = sy.sympify("sin(pi*x) * cos(pi*y) * exp(z)")
#    expr    = sy.sympify("exp(x)*exp(y)*exp(z)")
#    expr    = sy.sympify("1*x + 2*y + 3*z")
#    expr    = sy.sympify("0 + 1*x + 2*y + 3*z + 4*x**2 + 5*x*y + 6*y**2 + 7*y*z + 8*z**2 + 9*x*z + 10*x**3 + 11*x**2*y + 12*x*y**2 + 13*y**3 + 14*y**2*z + 15*y*z**2 + 16*z**3 + 17*z**2*x + 18*z*x**2 + 19*x*y*z")

    noise_eps = 0#1e-3 # introduce this much Gaussian noise into each sampled function value (use 0. to turn off)

    xi = np.array( (0.45, 0.25, 0.35) )  # point (x,y,z) where we wish to find the derivatives
#    xi = np.array( (0., 0., 0.) )  # point (x,y,z) where we wish to find the derivatives

    # Degree of the surrogate polynomial; a full polynomial of this order will be used.
    #
    # In the fit, when compared to the original function (if any is available),
    # usually the highest order will be nonsense, and the lower orders will be pretty accurate.
    #
    # (I.e. the unfittable part seems to favor the highest order; which OTOH has the highest spatial frequency. Maybe there's something here?)
    #
    fit_order = 4  # 0 (constant), 1 (linear), 2 (quadratic), 3 (cubic) or 4 (quartic)

#    weighting_method = wlsqm.WEIGHT_UNIFORM  # best overall fit for function values
    weighting_method = wlsqm.WEIGHT_CENTER  # emphasize center to improve derivatives at the point xi

    max_iter = 100  # maximum number of refinement iterations for iterative fitting

    do_sens = False  # do sensitivity analysis of solution? ( d( fi[j] ) / d( fk[k] ) )

    debug = False#True  # print row scaling and condition number information? (if True, then do_sens must be False; the combination with both True is not supported)

    # Bitmask of what we know at point xi. In this example, just set the bits;
    # the data (from expr) will be automatically inserted into fi[].
    #
    # See the constants b3_* in wlsqm.fitter.defs.
    #
    knowns = 1

    # How many neighbor points to generate (to simulate the meshless 'grid').
    #
    # At least n_unknowns  points are needed to make the model fitting work at all
    # (but then the fit will be nonsensical, since it is possible to make the polynomial
    #  pass through exactly those points).
    #
    # n_unknows + 1  is the first value that makes the fitting overdetermined,
    # i.e. where the least-squares procedure starts providing any advantage.
    #
    # Here "unknown" means any element of fi[] not tagged as known in the "knowns" bitmask.
    #
    nk = 200  # used if grid_type == 'random'

    r  = 1e-1  # neighborhood radius

#    grid_type = 'random'
    grid_type = 'stencil'
#    grid_type = 'sudoku'

    #########################
    # the test itself
    #########################

    print
    print "=" * 79
    print "3D case"
    print "=" * 79
    print

    print "expr: %s, xi = %s" % (expr, xi)

    labels = ["F",
              "DX", "DY", "DZ",
              "DX2", "DXDY", "DY2", "DYDZ", "DZ2", "DXDZ",
              "DX3", "DX2DY", "DXDY2", "DY3", "DY2DZ", "DYDZ2", "DZ3", "DXDZ2", "DX2DZ", "DXDYDZ",
              "DX4", "DX3DY", "DX2DY2", "DXDY3", "DY4", "DY3DZ", "DY2DZ2", "DYDZ3", "DZ4", "DXDZ3", "DX2DZ2", "DX3DZ", "DX2DYDZ", "DXDY2DZ", "DXDYDZ2" ]
    print "legend: %s" % ("\t".join(labels))
    knowns_str = ""
    for j in range(wlsqm.SIZE3):  # SIZE3 = maximum size of c matrix for 3D case
        if j > 0:
            knowns_str += '\t'
        if knowns & (1 << j):
            knowns_str += labels[j]
    print "knowns: %s" % knowns_str
#    # http://stackoverflow.com/questions/699866/python-int-to-binary
#    print "knowns (mask): %s" % format(knowns, '010b')[::-1]

    print "surrogate order: %d" % fit_order

    if noise_eps > 0.:
        print "simulating noisy input with eps = %g" % noise_eps

    # SymPy expr --> lambda(x,y)
    lambdify_numpy_3d = lambda expr: sy.lambdify(("x","y","z"), expr, modules="numpy")
    f          = lambdify_numpy_3d(expr)

    dfdx       = lambdify_numpy_3d(sy.diff(expr, "x"))
    dfdy       = lambdify_numpy_3d(sy.diff(expr, "y"))
    dfdz       = lambdify_numpy_3d(sy.diff(expr, "z"))

    d2fdx2     = lambdify_numpy_3d(sy.diff(expr, "x", 2))
    d2fdxdy    = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x"), "y" ))
    d2fdy2     = lambdify_numpy_3d(sy.diff(expr, "y", 2))
    d2fdydz    = lambdify_numpy_3d(sy.diff( sy.diff(expr, "y"), "z" ))
    d2fdz2     = lambdify_numpy_3d(sy.diff(expr, "z", 2))
    d2fdxdz    = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x"), "z" ))

    d3fdx3     = lambdify_numpy_3d(sy.diff(expr, "x", 3))
    d3fdx2dy   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x", 2), "y" ))
    d3fdxdy2   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x"), "y", 2 ))
    d3fdy3     = lambdify_numpy_3d(sy.diff(expr, "y", 3))
    d3fdy2dz   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "y", 2), "z" ))
    d3fdydz2   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "y"), "z", 2 ))
    d3fdz3     = lambdify_numpy_3d(sy.diff(expr, "z", 3))
    d3fdxdz2   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x"), "z", 2 ))
    d3fdx2dz   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x", 2), "z" ))
    d3fdxdydz  = lambdify_numpy_3d(sy.diff( sy.diff( sy.diff(expr, "x"), "y"), "z"))

    d4fdx4     = lambdify_numpy_3d(sy.diff(expr, "x", 4))
    d4fdx3dy   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x", 3), "y" ))
    d4fdx2dy2  = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x", 2), "y", 2 ))
    d4fdxdy3   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x"), "y", 3 ))
    d4fdy4     = lambdify_numpy_3d(sy.diff(expr, "y", 4))
    d4fdy3dz   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "y", 3), "z" ))
    d4fdy2dz2  = lambdify_numpy_3d(sy.diff( sy.diff(expr, "y", 2), "z" , 2))
    d4fdydz3   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "y"), "z", 3 ))
    d4fdz4     = lambdify_numpy_3d(sy.diff(expr, "z", 4))
    d4fdxdz3   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x"), "z", 3 ))
    d4fdx2dz2  = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x", 2), "z" , 2))
    d4fdx3dz   = lambdify_numpy_3d(sy.diff( sy.diff(expr, "x", 3), "z" ))
    d4fdx2dydz = lambdify_numpy_3d(sy.diff( sy.diff( sy.diff(expr, "x", 2), "y"), "z"))
    d4fdxdy2dz = lambdify_numpy_3d(sy.diff( sy.diff( sy.diff(expr, "x"), "y", 2), "z"))
    d4fdxdydz2 = lambdify_numpy_3d(sy.diff( sy.diff( sy.diff(expr, "x"), "y"), "z", 2))

    # list so we can refer to the functions by indices
    funcs = ( f,
              dfdx, dfdy, dfdz,
              d2fdx2, d2fdxdy, d2fdy2, d2fdydz, d2fdz2, d2fdxdz,
              d3fdx3, d3fdx2dy, d3fdxdy2, d3fdy3, d3fdy2dz, d3fdydz2, d3fdz3, d3fdxdz2, d3fdx2dz, d3fdxdydz,
              d4fdx4, d4fdx3dy, d4fdx2dy2, d4fdxdy3, d4fdy4, d4fdy3dz, d4fdy2dz2, d4fdydz3, d4fdz4, d4fdxdz3, d4fdx2dz2, d4fdx3dz, d4fdx2dydz, d4fdxdy2dz, d4fdxdydz2
            )

    # create neighbor points xk around the point xi - this simulates our meshless 'grid'
    #
    if grid_type == 'random':
        xk = np.tile(xi, (nk,1)) + r*2.*( np.random.sample( (nk,3) ) - 0.5 )

    elif grid_type == 'stencil':
        points_per_axis = max(1,fit_order) + 1

        tt = np.linspace(-1., 1., points_per_axis)
        X,Y,Z = np.meshgrid(tt,tt,tt)
        X = np.reshape(X, -1)
        Y = np.reshape(Y, -1)
        Z = np.reshape(Z, -1)

        # convert to list of (x,y) pairs, rejecting the point (0,0) (that represents xi itself), if present
        point_list = [ (x,y,z) for x,y,z in zip(X,Y,Z) if (x,y,z) != (0.,0.,0.) ]

        nk = len(point_list)
        xk = np.array( [ ( xi[0] + r*p[0], xi[1] + r*p[1], xi[2] + r*p[2] ) for p in point_list ] )

    elif grid_type == 'sudoku':
        points_per_axis = max(1,fit_order) + 1

        S,m = sudoku_lhs.sample(3, points_per_axis, 1)
        bins_per_axis = points_per_axis*m
        S = S / float(bins_per_axis - 1)  # scale the sample from [0, bins_per_axis-1]**3 to [0, 1]**3
        S = 2. * (S - 0.5) # move to [-1, 1]**3

        # If points_per_axis is odd, a bin exists exactly at the center, so Sudoku LHS may place one point there.
        #
        # This would coincide with the point xi, so it is not useful, because we want the neighbors to be
        # distinct from xi.
        #
        # Thus, for odd points_per_axis, filter the sample to remove the point at the origin if it happens to be there.
        # Note that because of the scaling, the coordinates might not be exactly zero. We HACK by checking numerical equality;
        # a proper solution would be to filter S before the conversion to float.
        #
        if points_per_axis % 2 == 1:
            point_list = S.tolist()
            oldlen = len(point_list)
            point_list = filter( lambda item: not (abs(item[0]) < 1e-8 and abs(item[1]) < 1e-8 and abs(item[2]) < 1e-8), point_list )
            S = np.array(point_list)
            if len(point_list) < oldlen:
                print "Sudoku LHS sampled the point at the origin; discarding it from the sample"

        nk = len(S)
        xk = np.tile(xi, (nk,1)) + r*S

    else:
        raise ValueError("Unknown grid_type '%s'; valid: 'random', 'stencil', 'sudoku'" % grid_type)

    # sample the function values at the neighbor points xk (these are used to fit the surrogate model)
    #
    sample_also_xi_str = " (and xi itself)" if knowns & 1 else ""
    print "sampling %d points%s" % (nk, sample_also_xi_str)
    fk = np.empty( (nk,), dtype=np.float64 )
    for k in xrange(nk):
        fk[k] = f( xk[k,0], xk[k,1], xk[k,2] )

    # simulate numerical errors by adding noise to the neighbor point function value samples
    #
    if noise_eps > 0.:
#        # uniform
#        noise = noise_eps*2.*(np.random.sample( np.shape(fk) ) - 0.5)

        # Gaussian, truncated
        mu    = 0.0
        sigma = noise_eps / 3.
        noise = np.random.normal( loc=mu, scale=sigma, size=np.shape(fk) )
        noise[noise < -3.*sigma] = -3.*sigma
        noise[noise > +3.*sigma] = +3.*sigma

        fk += noise

    # set knowns *at point xi*
    #
    # we use nan to spot unfilled entries
    fi = np.nan * np.empty( (wlsqm.SIZE3,), dtype=np.float64 )  # F, DX, DY, DZ, ... at point xi
    for d in range(wlsqm.SIZE3):
        if knowns & (1 << d):
            fi[d] = funcs[d]( xi[0], xi[1], xi[2] )  # fill in the known value  # TODO: add noise here too?

    # allocate array for sensitivity data
    #
    # for output; sens[k,j] = d(fi[j])/d(fk[k]) if f[i] unknown
    #                         nan if fi[j] known
    #
    # Note that if order=1, the part on second derivatives is not touched (so that an (nk,3) array
    # is valid); hence we pre-fill by nan.
    #
    if do_sens:
        sens = np.nan * np.empty( (nk,wlsqm.SIZE3), dtype=np.float64 )
    else:
        sens = None

    # fit the surrogate model (see wlsqm.fitter.simple for detailed documentation)
    #
    if debug:
        print  # blank line before debug info

    iterations_taken = wlsqm.fit_3D_iterative( xk, fk, xi, fi, sens, do_sens=do_sens, order=fit_order, knowns=knowns, debug=debug, weighting_method=weighting_method, max_iter=max_iter )
#    iterations_taken = wlsqm.fit_3D( xk, fk, xi, fi, sens, do_sens=do_sens, order=fit_order, knowns=knowns, debug=debug, weighting_method=weighting_method )

    print "refinement iterations taken: %d" % iterations_taken

    # check exact solution and relative error
    #
    exact = np.array( map( lambda func : func( xi[0], xi[1], xi[2] ), funcs ) )
    err   = (fi - exact)

    print
    print "derivatives at xi:"
    print "exact:"
    print exact
    print "wlsqm solution:"
    print fi
    if do_sens:
        print "sensitivity:"
        print sens
    print "abs error:"
    print err
    print "rel error:"
    print (err / exact)

    #########################
    # plotting
    #########################

    # surrogate model - the returned fi[] are actually the coefficients of a polynomial
    model    = wlsqm.lambdify_fit( xi, fi, dimension=3, order=fit_order )  # lambda x,y : ...

    print
    print "function values at neighbor points:"
    fxk =     f( xk[:,0], xk[:,1], xk[:,2] )
    mxk = model( xk[:,0], xk[:,1], xk[:,2] )
    print "exact:"
    print fxk
    print "wlsqm solution:"
    print mxk
    print "abs error:"
    errf = mxk - fxk
    print errf
    print "rel error:"
    print (errf / fxk)

    # comparison
    xx2        = np.linspace(xi[0] - r, xi[0] + r, nvis)
    yy2        = np.linspace(xi[1] - r, xi[1] + r, nvis)
    zz2        = np.linspace(xi[2] - r, xi[2] + r, nvis)
    X2,Y2,Z2   = np.meshgrid(xx2, yy2, zz2)
    W2         = model(X2,Y2,Z2)
    W3         =     f(X2,Y2,Z2)
    diff       = W2 - W3  # fitted - exact
    idx        = np.argmax(np.abs( diff ))
    diff_lin   = np.reshape(diff, -1)
    W3_lin     = np.reshape(W3, -1)
    maxerr_abs = diff_lin[idx]
    maxerr_rel = diff_lin[idx] / W3_lin[idx]
    print "largest absolute total fit error (over the domain of the fit, not just the neighbor points):"
    print "absolute: %g" % (maxerr_abs)
    print "relative: %g" % (maxerr_rel)


# one local model, 2D
#
def test2d():
    #########################
    # config
    #########################

    axislims = [0., 1., 0., 1.]  # [xmin, xmax, ymin, ymax], for plotting
    nvis = 101  # number of visualization points per axis

    # Let's manufacture a solution (for which we know the derivatives analytically):
    #
#    expr     = sy.sympify("2*x + 3*y")
#    expr     = sy.sympify("0.2*x + 0.3*y")
#    expr     = sy.sympify("1.0 + 2*x + 3*y + 4*x**2 + 5*x*y + 6*y**2")
#    expr     = sy.sympify("0.1 + 0.2*x + 0.3*y + 0.4*x**2 + 0.5*x*y + 0.6*y**2")
#    expr    = sy.sympify("sin(pi*x)")
    expr    = sy.sympify("sin(pi*x) * cos(pi*y)")
#    expr    = sy.sympify("exp(x) * 1/(1 + y) - 1")
#    expr    = sy.sympify("exp(x) * log(1 + y)")
#    expr     = sy.sympify("1.0 + 2*x + 3*y + 4*x**2 + 5*x*y + 6*y**2 + 7*x**3 + 8*y**4")

    noise_eps = 0#1e-3 # introduce this much Gaussian noise into each sampled function value (use 0. to turn off)

    xi = np.array( (0.45, 0.25) )  # point (x,y) where we wish to find the derivatives

    # Degree of the surrogate polynomial; a full polynomial of this order will be used.
    #
    # In the fit, when compared to the original function (if any is available),
    # usually the highest order will be nonsense, and the lower orders will be pretty accurate.
    #
    # (I.e. the unfittable part seems to favor the highest order; which OTOH has the highest spatial frequency. Maybe there's something here?)
    #
    fit_order = 4  # 0 (constant), 1 (linear), 2 (quadratic), 3 (cubic) or 4 (quartic)

#    weighting_method = wlsqm.WEIGHT_UNIFORM  # best overall fit for function values
    weighting_method = wlsqm.WEIGHT_CENTER  # emphasize center to improve derivatives at the point xi

    max_iter = 100  # maximum number of refinement iterations for iterative fitting

    do_sens = False  # do sensitivity analysis of solution? ( d( fi[j] ) / d( fk[k] ) )

    debug = False#True  # print row scaling and condition number information? (if True, then do_sens must be False; the combination with both True is not supported)

    # Bitmask of what we know at point xi. In this example, just set the bits;
    # the data (from expr) will be automatically inserted into fi[].
    #
    # Bits from least sig. to most sig.: F, DX, DY, DX2, DXDY, DY2, ... (see ordering of "labels", below)
    #
    knowns = 1

    # How many neighbor points to generate (to simulate the meshless 'grid').
    #
    # At least n_unknowns  points are needed to make the model fitting work at all
    # (but then the fit will be nonsensical, since it is possible to make the polynomial
    #  pass through exactly those points).
    #
    # n_unknows + 1  is the first value that makes the fitting overdetermined,
    # i.e. where the least-squares procedure starts providing any advantage.
    #
    # Here "unknown" means any element of fi[] not tagged as known in the "knowns" bitmask.
    #
    nk = 24  # used if grid_type == 'random'

    r  = 1e-1  # neighborhood radius

#    grid_type = 'random'
#    grid_type = 'stencil'
    grid_type = 'sudoku'

    #########################
    # the test itself
    #########################

    print
    print "=" * 79
    print "2D case"
    print "=" * 79
    print

    print "expr: %s, xi = %s" % (expr, xi)

    labels = ["F", "DX", "DY", "DX2", "DXDY", "DY2", "DX3", "DX2DY", "DXDY2", "DY3", "DX4", "DX3DY", "DX2DY2", "DXDY3", "DY4"]
    print "legend: %s" % ("\t".join(labels))
    knowns_str = ""
    for j in range(wlsqm.SIZE2):  # SIZE2 = maximum size of c matrix for 2D case
        if j > 0:
            knowns_str += '\t'
        if knowns & (1 << j):
            knowns_str += labels[j]
    print "knowns: %s" % knowns_str
#    # http://stackoverflow.com/questions/699866/python-int-to-binary
#    print "knowns (mask): %s" % format(knowns, '010b')[::-1]

    print "surrogate order: %d" % fit_order

    if noise_eps > 0.:
        print "simulating noisy input with eps = %g" % noise_eps

    # SymPy expr --> lambda(x,y)
    lambdify_numpy_2d = lambda expr: sy.lambdify(("x","y"), expr, modules="numpy")
    f         = lambdify_numpy_2d(expr)
    dfdx      = lambdify_numpy_2d(sy.diff(expr, "x"))
    dfdy      = lambdify_numpy_2d(sy.diff(expr, "y"))
    d2fdx2    = lambdify_numpy_2d(sy.diff(expr, "x", 2))
    d2fdxdy   = lambdify_numpy_2d(sy.diff( sy.diff(expr, "x"), "y" ))
    d2fdy2    = lambdify_numpy_2d(sy.diff(expr, "y", 2))
    d3fdx3    = lambdify_numpy_2d(sy.diff(expr, "x", 3))
    d3fdx2dy  = lambdify_numpy_2d(sy.diff( sy.diff(expr, "x", 2), "y" ))
    d3fdxdy2  = lambdify_numpy_2d(sy.diff( sy.diff(expr, "x"), "y", 2 ))
    d3fdy3    = lambdify_numpy_2d(sy.diff(expr, "y", 3))
    d4fdx4    = lambdify_numpy_2d(sy.diff(expr, "x", 4))
    d4fdx3dy  = lambdify_numpy_2d(sy.diff( sy.diff(expr, "x", 3), "y" ))
    d4fdx2dy2 = lambdify_numpy_2d(sy.diff( sy.diff(expr, "x", 2), "y", 2 ))
    d4fdxdy3  = lambdify_numpy_2d(sy.diff( sy.diff(expr, "x"), "y", 3 ))
    d4fdy4    = lambdify_numpy_2d(sy.diff(expr, "y", 4))
    funcs = (f, dfdx, dfdy, d2fdx2, d2fdxdy, d2fdy2, d3fdx3, d3fdx2dy, d3fdxdy2, d3fdy3, d4fdx4, d4fdx3dy, d4fdx2dy2, d4fdxdy3, d4fdy4)  # list so we can refer to the functions by indices

    # create neighbor points xk around the point xi - this simulates our meshless 'grid'
    #
    if grid_type == 'random':
        xk = np.tile(xi, (nk,1)) + r*2.*( np.random.sample( (nk,2) ) - 0.5 )

    elif grid_type == 'stencil':
        points_per_axis = max(1,fit_order) + 1

        tt = np.linspace(-1., 1., points_per_axis)
        X,Y = np.meshgrid(tt,tt)
        X = np.reshape(X, -1)
        Y = np.reshape(Y, -1)

        # convert to list of (x,y) pairs, rejecting the point (0,0) (that represents xi itself), if present
        point_list = [ (x,y) for x,y in zip(X,Y) if (x,y) != (0.,0.) ]

        nk = len(point_list)
        xk = np.array( [ ( xi[0] + r*p[0], xi[1] + r*p[1] ) for p in point_list ] )

    elif grid_type == 'sudoku':
        points_per_axis = max(1,fit_order) + 1

        S,m = sudoku_lhs.sample(2, points_per_axis, 1)
        bins_per_axis = points_per_axis*m
        S = S / float(bins_per_axis - 1)  # scale the sample from [0, bins_per_axis-1]**2 to [0, 1]**2
        S = 2. * (S - 0.5) # move to [-1, 1]**2

        # If points_per_axis is odd, a bin exists exactly at the center, so Sudoku LHS may place one point there.
        #
        # This would coincide with the point xi, so it is not useful, because we want the neighbors to be
        # distinct from xi.
        #
        # Thus, for odd points_per_axis, filter the sample to remove the point at the origin if it happens to be there.
        # Note that because of the scaling, the coordinates might not be exactly zero. We HACK by checking numerical equality;
        # a proper solution would be to filter S before the conversion to float.
        #
        if points_per_axis % 2 == 1:
            point_list = S.tolist()
            oldlen = len(point_list)
            point_list = filter( lambda item: not (abs(item[0]) < 1e-8 and abs(item[1]) < 1e-8), point_list )
            S = np.array(point_list)
            if len(point_list) < oldlen:
                print "Sudoku LHS sampled the point at the origin; discarding it from the sample"

        nk = len(S)
        xk = np.tile(xi, (nk,1)) + r*S

    else:
        raise ValueError("Unknown grid_type '%s'; valid: 'random', 'stencil', 'sudoku'" % grid_type)

    # sample the function values at the neighbor points xk (these are used to fit the surrogate model)
    #
    sample_also_xi_str = " (and xi itself)" if knowns & 1 else ""
    print "sampling %d points%s" % (nk, sample_also_xi_str)
    fk = np.empty( (nk,), dtype=np.float64 )
    for k in xrange(nk):
        fk[k] = f( xk[k,0], xk[k,1] )

    # simulate numerical errors by adding noise to the neighbor point function value samples
    #
    if noise_eps > 0.:
#        # uniform
#        noise = noise_eps*2.*(np.random.sample( np.shape(fk) ) - 0.5)

        # Gaussian, truncated
        mu    = 0.0
        sigma = noise_eps / 3.
        noise = np.random.normal( loc=mu, scale=sigma, size=np.shape(fk) )
        noise[noise < -3.*sigma] = -3.*sigma
        noise[noise > +3.*sigma] = +3.*sigma

        fk += noise

    # set knowns *at point xi*
    #
    # we use nan to spot unfilled entries
    fi = np.nan * np.empty( (wlsqm.SIZE2,), dtype=np.float64 )  # F, DX, DY, DX2, DXDY, DY2, DX3, DX2DY, DXDY2, DY3 at point xi
    for d in range(wlsqm.SIZE2):
        if knowns & (1 << d):
            fi[d] = funcs[d]( xi[0], xi[1] )  # fill in the known value  # TODO: add noise here too?

    # allocate array for sensitivity data
    #
    # for output; sens[k,j] = d(fi[j])/d(fk[k]) if f[i] unknown
    #                         nan if fi[j] known
    #
    # Note that if order=1, the part on second derivatives is not touched (so that an (nk,3) array
    # is valid); hence we pre-fill by nan.
    #
    if do_sens:
        sens = np.nan * np.empty( (nk,wlsqm.SIZE2), dtype=np.float64 )
    else:
        sens = None

    # fit the surrogate model (see wlsqm.fitter.simple for detailed documentation)
    #
    if debug:
        print  # blank line before debug info
    iterations_taken = wlsqm.fit_2D_iterative( xk, fk, xi, fi, sens, do_sens=do_sens, order=fit_order, knowns=knowns, debug=debug, weighting_method=weighting_method, max_iter=max_iter )
    print "refinement iterations taken: %d" % iterations_taken

    # check exact solution and relative error
    #
    exact = np.array( map( lambda func : func( xi[0], xi[1] ), funcs ) )
    err   = (fi - exact)

    print
    print "derivatives at xi:"
    print "exact:"
    print exact
    print "wlsqm solution:"
    print fi
    if do_sens:
        print "sensitivity:"
        print sens
    print "abs error:"
    print err
    print "rel error:"
    print (err / exact)

    #########################
    # plotting
    #########################

    xx   = np.linspace(axislims[0], axislims[1], nvis)
    yy   = np.linspace(axislims[2], axislims[3], nvis)
    X,Y  = np.meshgrid(xx, yy)
    W    = f(X,Y)

    # surrogate model - the returned fi[] are actually the coefficients of a polynomial
    model = wlsqm.lambdify_fit( xi, fi, dimension=2, order=fit_order )  # lambda x,y : ...
    xx2   = np.linspace(xi[0] - r, xi[0] + r, nvis)
    yy2   = np.linspace(xi[1] - r, xi[1] + r, nvis)
    X2,Y2 = np.meshgrid(xx2, yy2)
    W2    = model(X2,Y2)

#    # It is also possible to interpolate the model using the C API wrapper directly.
#    # The result is exactly the same; sometimes this API may be more convenient.
#    #
#    # Note that for the C API, the points x to which to interpolate the model must be formatted as x[k,:] = (xk,yk).
#    #
#    shp = np.shape(X2)
#    X2lin = np.reshape(X2, -1)
#    Y2lin = np.reshape(Y2, -1)
#    temp_x = np.array( [ (x,y) for x,y in zip(X2lin,Y2lin) ] )
#    out = wlsqm.interpolate_fit( xi, fi, dimension=2, order=fit_order, x=temp_x )
#    out = np.reshape( out, shp )
#    print
#    print "difference between Python and C API model interpolation:"
#    print out - W2  # should be close to zero

    print
    print "function values at neighbor points:"
    fxk = f( xk[:,0], xk[:,1] )
    mxk = model( xk[:,0], xk[:,1] )
    print "exact:"
    print fxk
    print "wlsqm solution:"
    print mxk
    print "abs error:"
    errf = mxk - fxk
    print errf
    print "rel error:"
    print (errf / fxk)

    # comparison
    W3   = f(X2,Y2)
    diff = W2 - W3  # fitted - exact
    idx  = np.argmax(np.abs( diff ))
    diff_lin = np.reshape(diff, -1)
    W3_lin   = np.reshape(W3, -1)
    maxerr_abs = diff_lin[idx]
    maxerr_rel = diff_lin[idx] / W3_lin[idx]
    print "largest absolute total fit error (over the domain of the fit, not just the neighbor points):"
    print "absolute: %g" % (maxerr_abs)
    print "relative: %g" % (maxerr_rel)

    fig = pl.figure(2, figsize=(12,6))  # for 2x1 subplots
#    fig = pl.figure(2, figsize=(12,12))
    fig.clf()

    ax = pl.subplot(1,2, 1)

    ax.plot( (xx[0],  xx[-1]), (yy[0],  yy[0]),  'k-' )
    ax.plot( (xx[-1], xx[-1]), (yy[0],  yy[-1]), 'k-' )
    ax.plot( (xx[0],  xx[-1]), (yy[-1], yy[-1]), 'k-' )
    ax.plot( (xx[0],  xx[0]),  (yy[0],  yy[-1]), 'k-' )

    ax.plot( xk[:,0], xk[:,1], linestyle='none', marker='o', markeredgecolor='r', markerfacecolor='none' )
    ax.plot( (xi[0] - r, xi[0] + r), (xi[1] - r, xi[1] - r), 'r-' )
    ax.plot( (xi[0] + r, xi[0] + r), (xi[1] - r, xi[1] + r), 'r-' )
    ax.plot( (xi[0] - r, xi[0] + r), (xi[1] + r, xi[1] + r), 'r-' )
    ax.plot( (xi[0] - r, xi[0] - r), (xi[1] - r, xi[1] + r), 'r-' )
    ax.plot( (xi[0],), (xi[1],), linestyle='none', marker='x', markeredgecolor='k', markerfacecolor='none' )
    pl.axis('tight')
    axis_marginize(ax, 0.02, 0.02)
    pl.grid(b=True, which='both')
    pl.xlabel('x')
    pl.ylabel('y')
    pl.subplot(1,2, 2)

    # make 3d plot of the function
    #
    # see http://matplotlib.sourceforge.net/examples/mplot3d/lines3d_demo.html

    # Axes3D has a tendency to underestimate how much space it needs; it draws its labels
    # outside the window area in certain orientations.
    #
    # This causes the labels to be clipped, which looks bad. We prevent this by creating the axes
    # in a slightly smaller rect (leaving a margin). This way the labels will show - outside the Axes3D,
    # but still inside the figure window.
    #
    # The final touch is to set the window background to a matching white, so that the
    # background of the figure appears uniform.
    #
    fig.patch.set_color( (1,1,1) )
    fig.patch.set_alpha( 1.0 )
    x0y0wh = [ 0.02, 0.02, 0.96, 0.96 ]  # left, bottom, width, height      (here as fraction of subplot area)

    # compute the corresponding figure coordinates for the 2x1 subplot layout
    x0y0wh[0] = 0.5 + 0.5*x0y0wh[0]  # left
    x0y0wh[2] = 0.5*x0y0wh[2]        # width

    ax = p3.Axes3D(fig, rect=x0y0wh)

    stride = max(1, (nvis-1)//10)  # pick a good-looking stride (for lines; we actually have more vertices, making a smoother-looking curve between the lines)
    # use linewidth=0 to remove the wireframe if desired.
#    surf = ax.plot_surface(X,Y,W, rstride=stride, cstride=stride, cmap=matplotlib.cm.Blues_r, clim=[fmin,fmax], linewidth=0.25, alpha=0.5)
    ax.plot_wireframe(X,Y,W, rstride=stride, cstride=stride, color='k', linewidth=0.5, linestyle='solid')
#        pl.colorbar(surf, shrink=0.5, aspect=5)
#        pl.colorbar(surf, shrink=0.96)

    # sampled points
    if noise_eps > 0.:
        ax.plot( xk[:,0], xk[:,1], f( xk[:,0], xk[:,1] ), linestyle='none', marker='o', markeredgecolor='k', markerfacecolor='none' )  # exact
        ax.plot( xk[:,0], xk[:,1], fk[:], linestyle='none', marker='o', markeredgecolor='r', markerfacecolor='none' )  # including noise
    else:
        ax.plot( xk[:,0], xk[:,1], f( xk[:,0], xk[:,1] ), linestyle='none', marker='o', markeredgecolor='r', markerfacecolor='none' )  # exact

    # surrogate model
    ax.plot_wireframe(X2,Y2,W2, rstride=stride, cstride=stride, color='r', linewidth=0.5, linestyle='solid')

    # point xi
    ax.plot( (xi[0],), (xi[1],), f( xi[0], xi[1] ), linestyle='none', marker='x', markeredgecolor='k', markerfacecolor='none' )

#    ax.view_init(20, -48)
#    ax.view_init(18, -46)
#    ax.view_init(18, -128)
    ax.view_init(34, 140)

    ax.axis('tight')
    ax.set_zlim(-1.01, 1.01)
    pl.xlabel('$x$')
    pl.ylabel('$y$')
    ax.set_title('f(x,y)')


# one local model, 1D
#
def test1d():
    #########################
    # config
    #########################

    axislims = [0., 1., 0., 1.]  # [xmin, xmax, ymin, ymax], for plotting
    nvis = 101  # number of visualization points

    # Let's manufacture a solution (for which we know the derivatives analytically):
    #
#    expr     = sy.sympify("2*x")
#    expr     = sy.sympify("0.2*x")
#    expr     = sy.sympify("1.0 + 2*x + 4*x**2")
#    expr     = sy.sympify("0.1 + 0.2*x + 0.4*x**2")
    expr     = sy.sympify("sin(pi*x)")
#    expr     = sy.sympify("1 / (1 + x)")
#    expr     = sy.sympify("exp(x)")
#    expr     = sy.sympify("log(1 + x)")
#    expr     = sy.sympify("1.0 + 2*x + 4*x**2 + 7*x**3")

    noise_eps = 0#1e-3  # introduce this much Gaussian noise into each sampled function value (use 0. to turn off)

    xi = 0.45  # point x where we wish to find the derivatives

    # Degree of the surrogate polynomial.
    #
    # In the fit, when compared to the original function (if any is available),
    # usually the highest order will be nonsense, and the lower orders will be pretty accurate.
    #
    fit_order = 4  # 0 (constant), 1 (linear), 2 (quadratic), 3 (cubic) or 4 (quartic)

#    weighting_method = wlsqm.WEIGHT_UNIFORM  # best overall fit for function values
    weighting_method = wlsqm.WEIGHT_CENTER  # emphasize center to improve derivatives at the point xi

    max_iter = 100  # maximum number of refinement iterations for iterative fitting

    do_sens = False  # do sensitivity analysis of solution? ( d( fi[j] ) / d( fk[k] ) )

    debug = False#True  # print row scaling and condition number information? (if True, then do_sens must be False; the combination with both True is not supported)

    # Bitmask of what we know at point xi. In this example, just set the bits;
    # the data (from expr) will be automatically inserted into fi[].
    #
    # Bits from least sig. to most sig.: F, DX, DX2, DX3, DX4
    #
    knowns = 1

    # How many neighbor points to generate (to simulate the meshless 'grid').
    #
    # At least n_unknowns  points are needed to make the model fitting work at all
    # (but then the fit will be nonsensical, since it is possible to make the polynomial
    #  pass through exactly those points).
    #
    # n_unknows + 1  is the first value that makes the fitting overdetermined,
    # i.e. where the least-squares procedure starts providing any advantage.
    #
    # Here "unknown" means any element of fi[] not tagged as known in the "knowns" bitmask.
    #
    nk = 7  # used if grid_type == 'random'

    r  = 1e-1  # neighborhood radius

#    grid_type = 'random'
    grid_type = 'stencil'

    #########################
    # the test itself
    #########################

    print
    print "=" * 79
    print "1D case"
    print "=" * 79
    print

    print "expr: %s, xi = %s" % (expr, xi)
    labels = ["F", "DX", "DX2", "DX3", "DX4"]
    print "legend: %s" % ("\t".join(labels))
    knowns_str = ""
    for j in range(wlsqm.SIZE1):  # SIZE1 = maximum size of c matrix for 1D case
        if j > 0:
            knowns_str += '\t'
        if knowns & (1 << j):
            knowns_str += labels[j]
    print "knowns: %s" % knowns_str

    print "surrogate order: %d" % fit_order

    if noise_eps > 0.:
        print "simulating noisy input with eps = %g" % noise_eps

    # SymPy expr --> lambda(x)
    lambdify_numpy_1d = lambda expr: sy.lambdify(("x"), expr, modules="numpy")
    f       = lambdify_numpy_1d(expr)
    dfdx    = lambdify_numpy_1d(sy.diff(expr, "x"))
    d2fdx2  = lambdify_numpy_1d(sy.diff(expr, "x", 2))
    d3fdx3  = lambdify_numpy_1d(sy.diff(expr, "x", 3))
    d4fdx4  = lambdify_numpy_1d(sy.diff(expr, "x", 4))
    funcs = (f, dfdx, d2fdx2, d3fdx3, d4fdx4)  # list so we can refer to the functions by indices

    # create neighbor points xk around the point xi - this simulates our meshless 'grid'
    #
    if grid_type == 'random':
        xk = xi + r*2.*( np.random.sample( (nk,) ) - 0.5 )

    elif grid_type == 'stencil':
        points_per_axis = max(1,fit_order) + 1

        tt = np.linspace(-1., 1., points_per_axis)

        # reject the point at the origin if it is there
        point_list = [ x for x in tt if x != 0. ]

        xk = np.array( [ xi + r*p for p in point_list ] )
        nk = len(xk)

    else:
        raise ValueError("Unknown grid_type '%s'; valid: 'random', 'stencil'" % grid_type)


    # sample the function values at the neighbor points xk (these are used to fit the surrogate model)
    #
    sample_also_xi_str = " (and xi itself)" if knowns & 1 else ""
    print "sampling %d points%s" % (nk, sample_also_xi_str)
    fk = np.empty( (nk,), dtype=np.float64 )
    for k in xrange(nk):
        fk[k] = f( xk[k] )

    # simulate numerical errors by adding noise to the neighbor point function value samples
    #
    if noise_eps > 0.:
#        # uniform
#        noise = noise_eps*2.*(np.random.sample( np.shape(fk) ) - 0.5)

        # Gaussian, truncated
        mu    = 0.0
        sigma = noise_eps / 3.
        noise = np.random.normal( loc=mu, scale=sigma, size=np.shape(fk) )
        noise[noise < -3.*sigma] = -3.*sigma
        noise[noise > +3.*sigma] = +3.*sigma

        fk += noise

    # set knowns *at point xi*
    #
    # we use nan to spot unfilled entries
    fi = np.nan * np.empty( (wlsqm.SIZE1,), dtype=np.float64 )
    for d in range(wlsqm.SIZE1):
        if knowns & (1 << d):
            fi[d] = funcs[d]( xi )  # fill in the known value  # TODO: add noise here too?

    # allocate array for sensitivity data
    #
    # for output; sens[k,j] = d(fi[j])/d(fk[k]) if f[i] unknown
    #                         nan if fi[j] known
    #
    # Note that if order=1, the part on second derivatives is not touched (so that an (nk,3) array
    # is valid); hence we pre-fill by nan.
    #
    if do_sens:
        sens = np.nan * np.empty( (nk,wlsqm.SIZE1), dtype=np.float64 )
    else:
        sens = None

    # do the numerical differentiation
    #
    # xk : in, (nk,) array of neighbor point coordinates
    # fk : in, (nk,) array of function values at the neighbor points
    # xi : in, double, coordinate of the point xi
    # fi : in/out: if order=2, (3,) array containing (f, dfdx, d2fdx2) at point xi
    #              if order=1, (2,) array containing (f, dfdx) at point xi
    #      on input:  those elements must be filled that correspond to the bitmask "knowns".
    #      on output: the unknown elements will be filled in (leaving the knowns untouched).
    # sens : out: if order=2, (nk,3) array containing sensitivity information.
    #             if order=1, (nk,2) array containing sensitivity information.
    #             if fi[j] is unknown: sens[k,j] = d( fi[j] ) / d( fk[k] )
    #             if fi[j] is known:   sens[:,j] = nan (to indicate "not applicable").
    # order  : in, order of the surrogate polynomial. Can be 1 or 2.
    #          Linear fit gives first derivatives only and has O(h**2) error.
    #          Quadratic fit gives first and second derivatives and has O(h**3) error.
    # knowns : in, bitmask describing what is known about the function at the point xi.
    #          See the b1_* (bitmask, 1D case) constants.
    #
    if debug:
        print  # blank line before debug info
    iterations_taken = wlsqm.fit_1D_iterative( xk, fk, xi, fi, sens, do_sens=do_sens, order=fit_order, knowns=knowns, debug=debug, weighting_method=weighting_method, max_iter=max_iter )
    print "refinement iterations taken: %d" % iterations_taken

    # check exact solution and relative error
    #
    exact = np.array( map( lambda func : func( xi ), funcs ) )
    err   = (fi - exact)

    print
    print "derivatives at xi:"
    print "exact:"
    print exact
    print "wlsqm solution:"
    print fi
    if do_sens:
        print "sensitivity:"
        print sens
    print "abs error:"
    print err
    print "rel error:"
    print (err / exact)

    #########################
    # plotting
    #########################

    nvis = 10001
    xx   = np.linspace(axislims[0], axislims[1], nvis)
    ww   = f(xx)

    # surrogate model - the returned fi[] are actually the coefficients of a polynomial
    model = wlsqm.lambdify_fit( xi, fi, dimension=1, order=fit_order )  # lambda x : ...
    xx2   = np.linspace(xi - r, xi + r, nvis)
    ww2   = model(xx2)

#    # It is also possible to interpolate the model using the C API wrapper directly.
#    # The result is exactly the same; sometimes this API may be more convenient.
#    #
#    # Note that for the C API, the points x to which to interpolate the model must be formatted as x[:] = (xk).
#    #
#    out = wlsqm.interpolate_fit( xi, fi, dimension=1, order=fit_order, x=xx2 )
#    print
#    print "difference between Python and C API model interpolation:"
#    print out - ww2  # should be close to zero

    print
    print "function values (and derivatives) at neighbor points:"
    flags = [ wlsqm.i1_F, wlsqm.i1_X, wlsqm.i1_X2, wlsqm.i1_X3, wlsqm.i1_X4 ]
    for label,func,flag in zip(labels,funcs,flags):
        m   = wlsqm.lambdify_fit( xi, fi, dimension=1, order=fit_order, diff=flag )  # using diff=..., derivatives of the model can be lambdified, too
        fxk = func( xk )
        mxk = m( xk )
        print label
        print "exact:"
        print fxk
        print "wlsqm solution:"
        print mxk
        print "abs error:"
        errf = mxk - fxk
        print errf
        print "rel error:"
        print (errf / fxk)

    # comparison
    ww3  = f(xx2)
    diff = ww2 - ww3  # fitted - exact
    idx  = np.argmax(np.abs( diff ))
    maxerr_abs = diff[idx]
    maxerr_rel = diff[idx] / ww3[idx]
    print "largest absolute total fit error (over the domain of the fit, not just the neighbor points):"
    print "absolute: %g" % (maxerr_abs)
    print "relative: %g" % (maxerr_rel)

    fig = pl.figure(1, figsize=(6,6))
    fig.clf()

    ax = pl.subplot(1,1, 1)

    # the function
    ax.plot( xx, ww, color='k', linewidth=0.5, linestyle='solid' )

    # surrogate model
    ax.plot( xx2, ww2, color='r', linewidth=1., linestyle='solid' )

    # sampled points
    #
    fxk = f(xk)
    if noise_eps > 0.:
        ax.plot( xk, fxk, linestyle='none', marker='o', markeredgecolor='k', markerfacecolor='none' )  # exact
        ax.plot( xk, fk, linestyle='none', marker='o', markeredgecolor='r', markerfacecolor='none' )  # including noise
    else:
        ax.plot( xk, fxk, linestyle='none', marker='o', markeredgecolor='r', markerfacecolor='none' )  # exact

    # helper lines for easier reading of figure
#    ax.plot( np.tile(xk, (2,1)), np.tile((-0.05, 0.05), (nk,1)).T, 'k--' )
    tmp = np.zeros( (2,nk), dtype=np.float64 )  # for vertical lines from zero level to f(xk)
    tmp[1,:] = fxk
    ax.plot( np.tile(xk, (2,1)), tmp, 'r--', linewidth=0.25 )

    # sampled region
    ax.plot( (xi - r, xi + r), (0., 0.), 'r-', linewidth=2. )
    ax.plot( (xi - r, xi - r), (0., f(xi-r)), 'r--', linewidth=0.5 )
    ax.plot( (xi + r, xi + r), (0., f(xi+r)), 'r--', linewidth=0.5 )

    # point xi
    ax.plot( xi, f(xi), linestyle='none', marker='x', markeredgecolor='k', markerfacecolor='none' )

    pl.axis('tight')
    axis_marginize(ax, 0.02, 0.02)
    pl.grid(b=True, which='both')
    pl.xlabel('x')
    pl.ylabel('y')


def main():
    test3d()
    test2d()
    test1d()
    testmany2d()
#    wlsqm.test_pointer_wrappers()
    print
    pl.show()

if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-
#
# WLSQM (Weighted Least SQuares Meshless): a fast and accurate meshless least-squares interpolator for Python, for scalar-valued data defined as point values on 1D, 2D and 3D point clouds.
#
# Low-level routines: distance matrix generation, problem matrix generation, solver.
#
# JJ 2016-11-30

# Set Cython compiler directives. This section must appear before any code!
#
# For available directives, see:
#
# http://docs.cython.org/en/latest/src/reference/compilation.html
#
# cython: wraparound  = False
# cython: boundscheck = False
# cython: cdivision   = True

from __future__ import division
from __future__ import absolute_import

from libc.stdlib cimport malloc, free
from libc.math cimport sqrt, log10
from libc.math cimport fabs as c_abs

from numpy import asanyarray
from numpy.linalg import cond  # only needed for debug mode

cimport wlsqm.utils.lapackdrivers as drivers

cimport wlsqm.fitter.defs   as defs    # C constants
cimport wlsqm.fitter.infra  as infra   # centralized memory allocation infrastructure
cimport wlsqm.fitter.interp as interp  # interpolation of fitted model

####################################################
# Distance matrix (c) generation
####################################################

# (Here lowercase "c" is the name of the distance matrix; has no relation to the C programming language.)

# Adapter that takes in parameters for 1D, 2D and 3D cases, calls the appropriate routine and passes on the relevant parameters.
#
# This is used to work around the static typing in C to allow a single dimension-agnostic implementation for the high-level fitting routines.
#
cdef void make_c_nD( infra.Case* case, double[::view.generic,::view.contiguous] xkManyD, double[::view.generic] xk1D ) nogil:
    if case.dimension == 3:
        make_c_3D( case, xkManyD )
    elif case.dimension == 2:
        make_c_2D( case, xkManyD )
    else: # case.dimension == 1:
        make_c_1D( case, xk1D )


# Fill the "c" array, describing inter-point distances as needed in model fitting - 3D case
#
# case     : in, case metadata and pointers to allocated memory
#     order    : in, the order of the polynomial to be fitted
#     no       : in, the number of DOFs in the unreduced system (corresponding to dimension=3 and order)
#     nk       : in, number of neighbors (data points used in fit)
#     c        : out, size (nk,no) array (C-contiguous!) of inter-point distances and their various powers as needed in the fitting
#     w        : out, size (nk,) array of fitting weights for each point xk, based on a function of their squared distances from xi
#     weighting_method : in, one of the constants WEIGHT_*. Specifies the type of weighting to use;
#                        different weightings are good for different use cases of WLSQM.
#     xi       : in, size (3,); the point xi itself
#
# xk       : in, size (nk,3); neighbor points; x[k,:] is the kth point
#
cdef void make_c_3D( infra.Case* case, double[::view.generic,::view.contiguous] xk ) nogil:

    # no-op in guest mode
    if not case.geometry_owned:
        return

    cdef double* c = case.c
    cdef double* w = case.w
    cdef int order = case.order
    cdef int weighting_method = case.weighting_method
    cdef int no    = case.no  # required size
    cdef int nk    = case.nk  # number of neighbors (data points used in fit) (usually the same as xk.shape[0])
    cdef double xi = case.xi
    cdef double yi = case.yi
    cdef double zi = case.zi

    cdef double d2, max_d2=0.  # euclidean distance squared

    cdef double dx,  dy,  dz   # x, y, z distances from the point xi
    cdef double dx2, dy2, dz2  # **2
    cdef double dx3, dy3, dz3  # **3

    # loop counters
    cdef int k       # neighbor points

    DEF onesixth = 1./6.
    DEF one24th  = 1./24.

    # generate the c^(j)_k array (see the documentation)
    #
    if order == 4:  # this is probably already pushing it (at double precision); may work for large enough region
        for k in range(nk):
            # Distances from point i to the neighbor points
            #
            dx  = xk[k,0] - xi  # signed x distance
            dy  = xk[k,1] - yi  # signed y distance
            dz  = xk[k,2] - zi  # signed z distance
            dx2 = dx*dx
            dy2 = dy*dy
            dz2 = dz*dz
            dx3 = dx2*dx
            dy3 = dy2*dy
            dz3 = dz2*dz

            d2 = dx2 + dy2 + dz2
            if d2 > max_d2:
                max_d2 = d2
            w[k] = d2  # store the squared distances for now, we'll convert them to weights in a second pass (once we have  max_d2  which we need anyway)

            # Distance coefficients (these multiply the derivatives in the Taylor series)
            #
            c[ k*no + defs.i3_F_c    ] = 1.

            c[ k*no + defs.i3_X_c    ] = dx
            c[ k*no + defs.i3_Y_c    ] = dy
            c[ k*no + defs.i3_Z_c    ] = dz

            c[ k*no + defs.i3_X2_c   ] = 0.5 * dx2
            c[ k*no + defs.i3_XY_c   ] = dx*dy
            c[ k*no + defs.i3_Y2_c   ] = 0.5 * dy2
            c[ k*no + defs.i3_YZ_c   ] = dy*dz
            c[ k*no + defs.i3_Z2_c   ] = 0.5 * dz2
            c[ k*no + defs.i3_XZ_c   ] = dx*dz

            c[ k*no + defs.i3_X3_c   ] = onesixth * dx3
            c[ k*no + defs.i3_X2Y_c  ] = 0.5 * dx2*dy
            c[ k*no + defs.i3_XY2_c  ] = 0.5 * dx*dy2
            c[ k*no + defs.i3_Y3_c   ] = onesixth * dy3
            c[ k*no + defs.i3_Y2Z_c  ] = 0.5 * dy2*dz
            c[ k*no + defs.i3_YZ2_c  ] = 0.5 * dy*dz2
            c[ k*no + defs.i3_Z3_c   ] = onesixth * dz3
            c[ k*no + defs.i3_XZ2_c  ] = 0.5 * dx*dz2
            c[ k*no + defs.i3_X2Z_c  ] = 0.5 * dx2*dz
            c[ k*no + defs.i3_XYZ_c  ] = dx*dy*dz

            c[ k*no + defs.i3_X4_c   ] = one24th  * dx2*dx2
            c[ k*no + defs.i3_X3Y_c  ] = onesixth * dx3*dy
            c[ k*no + defs.i3_X2Y2_c ] = 0.25     * dx2*dy2
            c[ k*no + defs.i3_XY3_c  ] = onesixth * dx*dy3
            c[ k*no + defs.i3_Y4_c   ] = one24th  * dy2*dy2
            c[ k*no + defs.i3_Y3Z_c  ] = onesixth * dy3*dz
            c[ k*no + defs.i3_Y2Z2_c ] = 0.25     * dy2*dz2
            c[ k*no + defs.i3_YZ3_c  ] = onesixth * dy*dz3
            c[ k*no + defs.i3_Z4_c   ] = one24th  * dz2*dz2
            c[ k*no + defs.i3_XZ3_c  ] = onesixth * dx*dz3
            c[ k*no + defs.i3_X2Z2_c ] = 0.25     * dx2*dz2
            c[ k*no + defs.i3_X3Z_c  ] = onesixth * dx3*dz
            c[ k*no + defs.i3_X2YZ_c ] = 0.5      * dx2*dy*dz
            c[ k*no + defs.i3_XY2Z_c ] = 0.5      * dx*dy2*dz
            c[ k*no + defs.i3_XYZ2_c ] = 0.5      * dx*dy*dz2

    elif order == 3:
        for k in range(nk):
            # Distances from point i to the neighbor points
            #
            dx  = xk[k,0] - xi  # signed x distance
            dy  = xk[k,1] - yi  # signed y distance
            dz  = xk[k,2] - zi  # signed z distance
            dx2 = dx*dx
            dy2 = dy*dy
            dz2 = dz*dz

            d2 = dx2 + dy2 + dz2
            if d2 > max_d2:
                max_d2 = d2
            w[k] = d2

            # Distance coefficients (these multiply the derivatives in the Taylor series)
            #
            c[ k*no + defs.i3_F_c   ] = 1.

            c[ k*no + defs.i3_X_c   ] = dx
            c[ k*no + defs.i3_Y_c   ] = dy
            c[ k*no + defs.i3_Z_c   ] = dz

            c[ k*no + defs.i3_X2_c  ] = 0.5 * dx2
            c[ k*no + defs.i3_XY_c  ] = dx*dy
            c[ k*no + defs.i3_Y2_c  ] = 0.5 * dy2
            c[ k*no + defs.i3_YZ_c  ] = dy*dz
            c[ k*no + defs.i3_Z2_c  ] = 0.5 * dz2
            c[ k*no + defs.i3_XZ_c  ] = dx*dz

            c[ k*no + defs.i3_X3_c  ] = onesixth * dx2*dx
            c[ k*no + defs.i3_X2Y_c ] = 0.5 * dx2*dy
            c[ k*no + defs.i3_XY2_c ] = 0.5 * dx*dy2
            c[ k*no + defs.i3_Y3_c  ] = onesixth * dy*dy2
            c[ k*no + defs.i3_Y2Z_c ] = 0.5 * dy2*dz
            c[ k*no + defs.i3_YZ2_c ] = 0.5 * dy*dz2
            c[ k*no + defs.i3_Z3_c  ] = onesixth * dz*dz2
            c[ k*no + defs.i3_XZ2_c ] = 0.5 * dx*dz2
            c[ k*no + defs.i3_X2Z_c ] = 0.5 * dx2*dz
            c[ k*no + defs.i3_XYZ_c ] = dx*dy*dz

    elif order == 2:
        for k in range(nk):
            # Distances from point i to the neighbor points
            #
            # Distances from point i to the neighbor points
            #
            dx  = xk[k,0] - xi  # signed x distance
            dy  = xk[k,1] - yi  # signed y distance
            dz  = xk[k,2] - zi  # signed z distance
            dx2 = dx*dx
            dy2 = dy*dy
            dz2 = dz*dz

            d2 = dx2 + dy2 + dz2
            if d2 > max_d2:
                max_d2 = d2
            w[k] = d2

            c[ k*no + defs.i3_F_c  ] = 1.

            c[ k*no + defs.i3_X_c  ] = dx
            c[ k*no + defs.i3_Y_c  ] = dy
            c[ k*no + defs.i3_Z_c  ] = dz

            c[ k*no + defs.i3_X2_c ] = 0.5 * dx2
            c[ k*no + defs.i3_XY_c ] = dx*dy
            c[ k*no + defs.i3_Y2_c ] = 0.5 * dy2
            c[ k*no + defs.i3_YZ_c ] = dy*dz
            c[ k*no + defs.i3_Z2_c ] = 0.5 * dz2
            c[ k*no + defs.i3_XZ_c ] = dx*dz

    elif order == 1:
        for k in range(nk):
            # Distances from point i to the neighbor points
            #
            dx  = xk[k,0] - xi  # signed x distance
            dy  = xk[k,1] - yi  # signed y distance
            dz  = xk[k,2] - zi  # signed z distance

            d2 = dx*dx + dy*dy + dz*dz
            if d2 > max_d2:
                max_d2 = d2
            w[k] = d2

            c[ k*no + defs.i3_F_c ] = 1.

            c[ k*no + defs.i3_X_c ] = dx
            c[ k*no + defs.i3_Y_c ] = dy
            c[ k*no + defs.i3_Z_c ] = dz

    else: # order == 0:
        # even in this case, we need the distances to compute the weights.
        for k in range(nk):
            # Distances from point i to the neighbor points
            #
            dx  = xk[k,0] - xi  # signed x distance
            dy  = xk[k,1] - yi  # signed y distance
            dz  = xk[k,2] - zi  # signed z distance

            d2 = dx*dx + dy*dy + dz*dz
            if d2 > max_d2:
                max_d2 = d2
            w[k] = d2

            c[ k*no + defs.i3_F_c ] = 1.

    # Convert squared distances to weights.
    #
    infra.Case_make_weights( case, max_d2 )


# Fill the "c" array, describing inter-point distances as needed in model fitting - 2D case
#
# case     : in, case metadata and pointers to allocated memory
#     order    : in, the order of the polynomial to be fitted
#     no       : in, the number of DOFs in the unreduced system (corresponding to dimension=2 and order)
#     nk       : in, number of neighbors (data points used in fit)
#     c        : out, size (nk,no) array (C-contiguous!) of inter-point distances and their various powers as needed in the fitting
#     w        : out, size (nk,) array of fitting weights for each point xk, based on a function of their squared distances from xi
#     weighting_method : in, one of the constants WEIGHT_*. Specifies the type of weighting to use;
#                        different weightings are good for different use cases of WLSQM.
#     xi       : in, size (2,); the point xi itself
#
# xk       : in, size (nk,2); neighbor points; x[k,:] is the kth point
#
cdef void make_c_2D( infra.Case* case, double[::view.generic,::view.contiguous] xk ) nogil:

    # no-op in guest mode
    if not case.geometry_owned:
        return

    cdef double* c = case.c
    cdef double* w = case.w
    cdef int order = case.order
    cdef int weighting_method = case.weighting_method
    cdef int no    = case.no  # required size
    cdef int nk    = case.nk  # number of neighbors (data points used in fit) (usually the same as xk.shape[0])
    cdef double xi = case.xi
    cdef double yi = case.yi

    cdef double d2, max_d2=0.  # euclidean distance squared

    cdef double dx,  dy   # x, y distances from the point xi
    cdef double dx2, dy2  # **2
    cdef double dx3, dy3  # **3

    # loop counters
    cdef int k       # neighbor points

    DEF onesixth = 1./6.
    DEF one24th  = 1./24.

    # generate the c^(j)_k array (see the documentation)
    #
    if order == 4:  # this is probably already pushing it (at double precision); may work for large enough region
        for k in range(nk):
            # Distances from point i to the neighbor points
            #
            dx = xk[k,0] - xi  # signed x distance
            dy = xk[k,1] - yi  # signed y distance
            dx2 = dx*dx
            dy2 = dy*dy
            dx3 = dx2*dx
            dy3 = dy2*dy

            d2 = dx2 + dy2
            if d2 > max_d2:
                max_d2 = d2
            w[k] = d2  # store the squared distances for now, we'll convert them to weights in a second pass (once we have  max_d2  which we need anyway)

            # Distance coefficients (these multiply the derivatives in the Taylor series)
            #
            c[ k*no + defs.i2_F_c    ] = 1.         # c^(0)_k  in the documentation

            c[ k*no + defs.i2_X_c    ] = dx         # c^(1)_k
            c[ k*no + defs.i2_Y_c    ] = dy         # c^(2)_k

            c[ k*no + defs.i2_X2_c   ] = 0.5 * dx2  # c^(3)_k
            c[ k*no + defs.i2_XY_c   ] = dx*dy      # c^(4)_k
            c[ k*no + defs.i2_Y2_c   ] = 0.5 * dy2  # c^(5)_k

            c[ k*no + defs.i2_X3_c   ] = onesixth * dx3
            c[ k*no + defs.i2_X2Y_c  ] = 0.5 * dx2*dy
            c[ k*no + defs.i2_XY2_c  ] = 0.5 * dx*dy2
            c[ k*no + defs.i2_Y3_c   ] = onesixth * dy3

            c[ k*no + defs.i2_X4_c   ] = one24th  * dx2*dx2
            c[ k*no + defs.i2_X3Y_c  ] = onesixth * dx3*dy
            c[ k*no + defs.i2_X2Y2_c ] = 0.25     * dx2*dy2
            c[ k*no + defs.i2_XY3_c  ] = onesixth * dx*dy3
            c[ k*no + defs.i2_Y4_c   ] = one24th  * dy2*dy2

    elif order == 3:
        for k in range(nk):
            # Distances from point i to the neighbor points
            #
            dx = xk[k,0] - xi  # signed x distance
            dy = xk[k,1] - yi  # signed y distance
            dx2 = dx*dx
            dy2 = dy*dy

            d2 = dx2 + dy2
            if d2 > max_d2:
                max_d2 = d2
            w[k] = d2

            # Distance coefficients (these multiply the derivatives in the Taylor series)
            #
            c[ k*no + defs.i2_F_c   ] = 1.         # c^(0)_k  in the documentation

            c[ k*no + defs.i2_X_c   ] = dx         # c^(1)_k
            c[ k*no + defs.i2_Y_c   ] = dy         # c^(2)_k

            c[ k*no + defs.i2_X2_c  ] = 0.5 * dx2  # c^(3)_k
            c[ k*no + defs.i2_XY_c  ] = dx*dy      # c^(4)_k
            c[ k*no + defs.i2_Y2_c  ] = 0.5 * dy2  # c^(5)_k

            c[ k*no + defs.i2_X3_c  ] = onesixth * dx2*dx
            c[ k*no + defs.i2_X2Y_c ] = 0.5 * dx2*dy
            c[ k*no + defs.i2_XY2_c ] = 0.5 * dx*dy2
            c[ k*no + defs.i2_Y3_c  ] = onesixth * dy*dy2

    elif order == 2:
        for k in range(nk):
            dx = xk[k,0] - xi  # signed x distance
            dy = xk[k,1] - yi  # signed y distance
            dx2 = dx*dx
            dy2 = dy*dy

            d2 = dx2 + dy2
            if d2 > max_d2:
                max_d2 = d2
            w[k] = d2

            c[ k*no + defs.i2_F_c  ] = 1.         # c^(0)_k  in the documentation

            c[ k*no + defs.i2_X_c  ] = dx         # c^(1)_k
            c[ k*no + defs.i2_Y_c  ] = dy         # c^(2)_k

            c[ k*no + defs.i2_X2_c ] = 0.5 * dx2  # c^(3)_k
            c[ k*no + defs.i2_XY_c ] = dx*dy      # c^(4)_k
            c[ k*no + defs.i2_Y2_c ] = 0.5 * dy2  # c^(5)_k

    elif order == 1:
        for k in range(nk):
            dx = xk[k,0] - xi  # signed x distance
            dy = xk[k,1] - yi  # signed y distance

            d2 = dx*dx + dy*dy
            if d2 > max_d2:
                max_d2 = d2
            w[k] = d2

            c[ k*no + defs.i2_F_c ] = 1.         # c^(0)_k  in the documentation

            c[ k*no + defs.i2_X_c ] = dx         # c^(1)_k
            c[ k*no + defs.i2_Y_c ] = dy         # c^(2)_k

    else: # order == 0:
        # even in this case, we need the distances to compute the weights.
        for k in range(nk):
            dx = xk[k,0] - xi  # signed x distance
            dy = xk[k,1] - yi  # signed y distance

            d2 = dx*dx + dy*dy
            if d2 > max_d2:
                max_d2 = d2
            w[k] = d2

            c[ k*no + defs.i2_F_c ] = 1.         # c^(0)_k  in the documentation

    # Convert squared distances to weights.
    #
    infra.Case_make_weights( case, max_d2 )


# Fill the "c" array, describing inter-point distances as needed in model fitting - 1D case
#
# case     : in, case metadata and pointers to allocated memory
#     order    : in, the order of the polynomial to be fitted
#     no       : in, the number of DOFs in the unreduced system (corresponding to dimension=1 and order)
#     nk       : in, number of neighbors (data points used in fit)
#     c        : out, size (nk,no) array (C-contiguous!) of inter-point distances and their various powers as needed in the fitting
#     w        : out, size (nk,) array of fitting weights for each point xk, based on a function of their squared distances from xi
#     weighting_method : in, one of the constants WEIGHT_*. Specifies the type of weighting to use;
#                        different weightings are good for different use cases of WLSQM.
#     xi       : in, the point xi itself
#
# xk       : in, size (nk,); neighbor points; x[k] is the kth point
#
cdef void make_c_1D( infra.Case* case, double[::view.generic] xk ) nogil:

    # no-op in guest mode
    if not case.geometry_owned:
        return

    cdef double* c = case.c
    cdef double* w = case.w
    cdef int order = case.order
    cdef int weighting_method = case.weighting_method
    cdef int no    = case.no  # required size
    cdef int nk    = case.nk  # number of neighbors (data points used in fit) (usually the same as xk.shape[0])
    cdef double xi = case.xi

    cdef double max_d2=0.  # euclidean distance squared (largest seen)
    cdef double dx, dx2

    # loop counters
    cdef int k       # neighbor points

    DEF onesixth = 1./6.
    DEF one24th  = 1./24.

    # generate the c^(j)_k array (see the documentation)
    #
    if order == 4:  # this is probably already pushing it (at double precision); may work for large enough region
        for k in range(nk):
            # Distances from point i to the neighbor points
            #
            dx  = xk[k] - xi  # signed x distance
            dx2 = dx*dx

            if dx2 > max_d2:
                max_d2 = dx2
            w[k] = dx2  # store the squared distances for now, we'll convert them to weights in a second pass (once we have  max_d2  which we need anyway)

            # Distance coefficients (these multiply the derivatives in the Taylor series)
            #
            c[ k*no + defs.i1_F_c  ] = 1.
            c[ k*no + defs.i1_X_c  ] = dx
            c[ k*no + defs.i1_X2_c ] = 0.5 * dx2
            c[ k*no + defs.i1_X3_c ] = onesixth * dx*dx2
            c[ k*no + defs.i1_X4_c ] = one24th  * dx2*dx2

    elif order == 3:
        for k in range(nk):
            dx  = xk[k] - xi
            dx2 = dx*dx

            if dx2 > max_d2:
                max_d2 = dx2
            w[k] = dx2

            c[ k*no + defs.i1_F_c  ] = 1.
            c[ k*no + defs.i1_X_c  ] = dx
            c[ k*no + defs.i1_X2_c ] = 0.5 * dx2
            c[ k*no + defs.i1_X3_c ] = onesixth * dx*dx2

    elif order == 2:
        for k in range(nk):
            dx  = xk[k] - xi
            dx2 = dx*dx

            if dx2 > max_d2:
                max_d2 = dx2
            w[k] = dx2

            c[ k*no + defs.i1_F_c  ] = 1.
            c[ k*no + defs.i1_X_c  ] = dx
            c[ k*no + defs.i1_X2_c ] = 0.5 * dx2

    elif order == 1:
        for k in range(nk):
            dx  = xk[k] - xi
            dx2 = dx*dx

            if dx2 > max_d2:
                max_d2 = dx2
            w[k] = dx2

            c[ k*no + defs.i1_F_c ] = 1.
            c[ k*no + defs.i1_X_c ] = dx

    else: # order == 0:
        # even in this case, we need the distances to compute the weights.
        for k in range(nk):
            dx  = xk[k] - xi
            dx2 = dx*dx

            if dx2 > max_d2:
                max_d2 = dx2
            w[k] = dx2

            c[ k*no + defs.i1_F_c ] = 1.

    # Convert squared distances to weights.
    #
    infra.Case_make_weights( case, max_d2 )


####################################################
# Problem matrix (A) generation
####################################################

# These routines generate the problem matrix A, based on the distance matrix C.

# Create the problem matrix A (naively, without scaling - see also preprocess_A()).
#
# case   : in, case metadata and pointers to allocated memory
#     A      : out, (nr, nr), Fortran-contiguous, the matrix A
#     c      : in, (nk,no) coefficient array, as output from make_c_?D()
#     w      : in, (nk,) array of weights, as output from make_c_?D()
#     nk     : in, number of neighbors (data points used in fit)
#     no     : in, number of DOFs in the original unreduced system, as output from make_c_?D()
#     nr     : in, number of DOFs in the reduced system, as output from remap()
#     knowns : in, knowns bitmask
#     o2r    : out, array of size (no,), will contain DOF mapping original --> reduced   (must be allocated by caller!)
#     r2o    : out, array of size (no,), will contain DOF mapping reduced  --> original  (must be allocated by caller!)
#
cdef void make_A( infra.Case* case ) nogil:

    # no-op in guest mode
    if not case.geometry_owned:
        return

    # all tagged as knowns --> nothing to solve
    #
    if case.nr < 1:
        return  # can't use GIL and raise exception, since we'll be called in the OpenMP threads

    cdef double* A        = case.A  # (nr, nr), Fortran-contiguous
    cdef double* c        = case.c
    cdef double* w        = case.w
    cdef int nk           = case.nk
    cdef int nr           = case.nr
    cdef int no           = case.no
    cdef long long knowns = case.knowns
    cdef int* o2r         = case.o2r
    cdef int* r2o         = case.r2o

    # construct the reduced system (naively, without scaling; preconditioning is applied later in preprocess_A())
    #
    cdef double acc           # accumulator for summation
    cdef int j, oj, m, om, k  # loop counters
    for j in range(nr):      # row in reduced system
        oj = r2o[j]  # index of reduced DOF j in original system

        # fill reduced matrix A with the coefficients of the reduced part
        #
        for m in range(nr):  # col in reduced system
            om = r2o[m]  # index of reduced DOF m in original system
            acc = 0.
            for k in range(nk):
                # c is indexed using indices of the original system (and is C-contiguous)
                acc += w[k] * c[ k*no + om ] * c[ k*no + oj ]  # apply w[k], the fitting weight of the error component k
            A[j + nr*m]     = acc  # A is indexed using indices of the reduced system (and is Fortran-contiguous)

# Precondition (scale) and LU factorize A.
#
# case      : in, case metadata and pointers to allocated memory
#     A         : in/out. Upon enter, the original matrix A as output by make_A().
#                         Upon exit,  the (packed) LU factorization of the preconditioned A.
#     nr        : in, number of DOFs in the reduced system. ("A" is of size (nr,nr))
#     row_scale : out, (nr,) array of row scaling factors  (must be allocated by caller!)
#                      When solving, this is needed to scale the RHS accordingly.
#     col_scale : out, (nr,) array of column scaling factors  (must be allocated by caller!)
#                      When solving, this is needed to scale the solution.
#     o2r       : in, array of size (no,), DOF mapping original --> reduced, as output by make_A()   (only used if debug=True)
#     r2o       : in, array of size (no,), DOF mapping reduced  --> original, as output by make_A()  (only used if debug=True)
#
# debug     : in, boolean. If debug is True, compute the condition number (2-norm directly via SVD) of the original and scaled A.
#             The condition numbers are written in case.cond_orig and case.cond_scaled.
#
cdef void preprocess_A( infra.Case* case, int debug ) nogil:

    # no-op in guest mode
    if not case.geometry_owned:
        return

    cdef double* A         = case.A
    cdef int* ipiv         = case.ipiv
    cdef double* row_scale = case.row_scale
    cdef double* col_scale = case.col_scale
    cdef int nr            = case.nr
    cdef int* o2r          = case.o2r
    cdef int* r2o          = case.r2o

    # all tagged as knowns --> nothing to solve
    #
    if nr < 1:
        return  # can't use GIL and raise exception, since we'll be called in the OpenMP threads

    # We may need a copy of A for computing condition numbers.
    #
    cdef double* Acopy = <double*>0
    if debug:
        Acopy = <double*>malloc( nr*nr*sizeof(double) )  # bypass the custom allocator; debug mode is not used in production code where memory fragmentation may matter
        drivers.copygeneral_c( Acopy, A, nr, nr )

    # Use scaling to improve the condition number of A, resulting in more correct digits in the solution.
    # The effects are drastic especially in high-order fitting.
    #
    # Specifically, we use DGESV, which works by LU factorization with pivoting. For this algorithm the
    # relative accuracy is O(kappa(A) * machine_epsilon), where at double precision, machine_epsilon ~ 1e-16.
    # ( http://scicomp.stackexchange.com/questions/19289/are-direct-solvers-affect-by-the-condition-number-of-a-matrix )
    #
    cdef int iterations_taken = drivers.rescale_ruiz2001_c( A, nr, nr, row_scale, col_scale )  # compute row and column scaling factors
    drivers.apply_scaling_c( A, nr, nr, row_scale, col_scale )     # using the computed factors, scale A in-place

    # Compute the condition numbers before factorizing A.
    #
    # We can't use dgecon() (which would use the factorized A), because we want the 2-norm condition number and dgecon() only supports 1-norm and infinity-norm.
    #
    cdef double* S = <double*>0
    cdef double cond_orig, cond_scaled
    cdef int j
    if debug:
        # Estimate the condition number (before and after scaling) and save it to the Case instance.
        #
        # According to https://docs.scipy.org/doc/numpy/reference/generated/numpy.linalg.cond.html
        # NumPy's cond() computes the 2-norm condition number (default) directly from the SVD.
        # http://mathworld.wolfram.com/ConditionNumber.html says this is max(S)/min(S).
        #
        # We can use LAPACK's dgesvd() to do this at the C level.
        # The output is sorted so that S[i] >= S[i+1].

        # original matrix
        S = <double*>malloc( nr*sizeof(double) )  # bypass the custom allocator; debug mode is not used in production code where memory fragmentation may matter
        drivers.svd_c( Acopy, nr, nr, S )  # result goes into S; Acopy is destroyed (overwritten)
        case.cond_orig   = S[0] / S[nr-1]

        # scaled matrix
        drivers.copygeneral_c( Acopy, A, nr, nr )
        drivers.svd_c( Acopy, nr, nr, S )
        case.cond_scaled = S[0] / S[nr-1]

        free( <void*>S )

    # Compute LU factorization of scaled A.
    #
    drivers.generalfactor_c( A, ipiv, nr )  # TODO: error handling (can't use GIL and catch exception, since we'll be called in the OpenMP threads)

    if debug:
        free( <void*>Acopy )  # bypassing the custom allocator, stuff for debug mode only


####################################################
# RHS handling and solving
####################################################

# Perform model fitting for given data.
#
# case      : in, case metadata and pointers to allocated memory
#     no        : in, number of DOFs in the original (unreduced) system, as output by make_c_?D()
#     c         : coefficient array of size (nk,no), as output by make_c_?D()
#     w         : in, (nk,) array of weights, as output from make_c_?D()
#
#     nr        : in, number of DOFs in the reduced system. ("A" is of size (nr,nr)), as output by make_A()
#     o2r       : in, array of size (no,), DOF mapping original --> reduced, as output by make_A()
#     r2o       : in, array of size (no,), DOF mapping reduced  --> original, as output by make_A()
#
#     A         : in, the (packed) LU factorization of the preconditioned A, as output by preprocess_A().
#     ipiv      : in, pivot information for solver, as output by preprocess_A().
#     row_scale : in, row scaling factors for scaling RHS, as output by preprocess_A().
#     col_scale : in, column scaling factors for scaling solution, as output by preprocess_A().
#
#     fi        : in/out, passed through from fit_*(), function data at xi; knowns in, unknowns out
#                 The number of elements in fi is taken from case.no.
#
#     knowns    : in, passed through from fit_*()
#
#     wrk       : work space, must be allocated by caller. (done by Case_new())
#                 If do_sens=False, must have space for (nr,) doubles.
#                 If do_sens=True,  must have space for (nr,) + (nr,nk) = (nk+1)*nr doubles.
#
# fk        : in, passed through from fit_*(), function value data at the neighbor points
#
# sens      : out, passed through from fit_*()
# do_sens   : in, boolean: whether to perform sensitivity analysis. If False, "sens" can be none.
#             (if the Case object has its do_sens flag set, it *supports* sensitivity analysis;
#              this flag determines whether we actually *do* it.)
#
# taskid    : in, number of parallel processing task (0, 1, ..., ntasks-1)
#             Used to get access to the RHS work space.
#
cdef void solve( infra.Case* case, double[::view.generic] fk, double[::view.generic,::view.contiguous] sens, int do_sens, int taskid ) nogil:

    cdef double nan = 0./0.  # FIXME: a better way to get NaN than to abuse IEEE-754 specification?

    cdef int no            = case.no
    cdef int nr            = case.nr
    cdef int nk            = case.nk  # usually the same as fk.shape[0]  # number of neighbors (data points used in fit)

    # all tagged as knowns --> nothing to solve (important; LAPACK doesn't like N = 0)
    #
    if nr < 1:
        return

    cdef long long knowns  = case.knowns

    cdef double* c         = case.c
    cdef double* w         = case.w
    cdef int* o2r          = case.o2r
    cdef int* r2o          = case.r2o

    cdef double* LU        = case.A  # case.A has already been factored when solve() is called
    cdef int* ipiv         = case.ipiv
    cdef double* row_scale = case.row_scale
    cdef double* col_scale = case.col_scale

    cdef double* fi        = case.fi

    cdef double* wrk       = infra.Case_get_wrk( case, taskid )

    cdef double* b = &wrk[0]   # (nr,)
    cdef double* s = &wrk[nr]  # the rest of the work space; (nr, nk), Fortran-contiguous

    # construct the RHS of the reduced system, applying also row scaling
    #
    cdef double acc    # accumulator for summation
    cdef int j, oj, k  # loop counters
    if do_sens:
        for j in range(nr):      # row in reduced system
            oj = r2o[j]  # index of reduced DOF j in original system

            acc = 0.
            for k in range(nk):
                acc += w[k] * fk[k] * c[ k*no + oj ]  # apply also w[k], the fitting weight of error component k

                # sensitivity of the solution w.r.t. fk[k]
                # (only one term in each row; this is d(b) / d( fk[k] ), leaving only c^(j)_k)
                s[j + nr*k] = row_scale[j] * w[k] * c[ k*no + oj ]
            b[j] = row_scale[j] * acc
    else:
        for j in range(nr):      # row in reduced system
            oj = r2o[j]  # index of reduced DOF j in original system

            acc = 0.
            for k in range(nk):
                acc += w[k] * fk[k] * c[ k*no + oj ]   # apply also w[k], the fitting weight of error component k
            b[j] = row_scale[j] * acc

    # apply algebraic elimination to the knowns
    #
    cdef int om
    for om in range(no):  # DOF in original system
        # handle only the eliminated DOFs
        if knowns & (1LL << om):
            # om now represents an eliminated column of the original system

            # update all rows of b in the reduced system.
            #
            # We must account for row scaling, because the whole row is scaled, and hence b, too.
            #
            # But column scaling cancels out for eliminated DOFs:
            #
            #   fi[om] * c[ k*SIZE + om ] = (fi[om] * col_scale) * (c[ k*SIZE + om ] / col_scale)
            #   ======   ================   --------------------   ------------------------------
            #   "small"    original coeff   "big" (scaled) units   scaled (big) coeff in A
            #   (orig.)
            #    units
            #
            # or intuitively:
            #
            #   many small units * small coeff = few big units * scaled (big) coeff
            #
            # This is fortunate, because the column scaling factor has not been computed for the eliminated columns!
            #
            for j in range(nr):  # row in reduced system
                oj = r2o[j]  # index of reduced DOF j in original system
                for k in range(nk):
                    b[j] -= fi[om] * w[k] * c[ k*no + om ] * c[ k*no + oj ] * row_scale[j]   # apply also w[k], the fitting weight of error component k

            # also, we may as well do this here:
            if do_sens:  # no is rather small, not many extra instructions
                for k in range(nk):
                    sens[k, om] = nan  # no sensitivity data for eliminated DOFs

    # solve (overwriting b with solution)
    #
    drivers.generalfactored_c( LU, ipiv, b, nr )

    # perform sensitivity analysis (overwriting s with the sensitivity data)
    #
    if do_sens:  # if no sensitivity analysis, major runtime savings occur here
        # we must loop as there is no "generalfactoreds_c()".
        for k in range(nk):
            drivers.generalfactored_c( LU, ipiv, &s[k*nr], nr )

    # extract solution (filling in only the unknowns in fi[])
    #
    for j in range(nr):  # row in reduced system
        oj = r2o[j]  # index of reduced DOF j in original system

        fi[oj] = b[j] * col_scale[j]  # compute original x from the column-scaled x

        # fill sensitivity data w.r.t. each unknown j and each fk[k]
        if do_sens:  # again, small nr, not many extra instructions
            for k in range(nk):
                sens[k, oj] = s[nr*k + j] * col_scale[j]


# TODO/FIXME: This function is identical to solve() except the signature. We do this senseless code duplication because solve_iterative() needs a version with raw pointers to contiguous C-level buffers to avoid GIL.
#
# TODO/FIXME: Memoryviews can be *accessed* without the GIL; the problem is that *creating* memoryview objects requires the GIL. The function solve_iterative() can't afford to do that,
# TODO/FIXME: because it is a low-level function that may be (and actually is; see the .pyx source for wlsqm.fitter.expert) called from an OpenMP parallel loop with the GIL released.
#
# Note that this version allows one to use a separate array fi, which is not necessarily the one stored in the Case instance. solve_iterative() uses this feature to compute the iterative updates.
#
# This function is not exported.
#
# - the number of elements in fk is taken from case.nk
# - the number of elements in fi is taken from case.no
#
cdef void solve_contig( infra.Case* case, double* fk, double* fi, double[::view.generic,::view.contiguous] sens, int do_sens, int taskid ) nogil:

    cdef double nan = 0./0.  # FIXME: a better way to get NaN than to abuse IEEE-754 specification?

    cdef int no            = case.no
    cdef int nr            = case.nr
    cdef int nk            = case.nk  # usually the same as fk.shape[0]  # number of neighbors (data points used in fit)

    # all tagged as knowns --> nothing to solve (important; LAPACK doesn't like N = 0)
    #
    if nr < 1:
        return

    cdef long long knowns  = case.knowns

    cdef double* c         = case.c
    cdef double* w         = case.w
    cdef int* o2r          = case.o2r
    cdef int* r2o          = case.r2o

    cdef double* LU        = case.A  # case.A has already been factored when solve() is called
    cdef int* ipiv         = case.ipiv
    cdef double* row_scale = case.row_scale
    cdef double* col_scale = case.col_scale

    cdef double* wrk       = infra.Case_get_wrk( case, taskid )

    cdef double* b = &wrk[0]   # (nr,)
    cdef double* s = &wrk[nr]  # the rest of the work space; (nr, nk), Fortran-contiguous

    # construct the RHS of the reduced system, applying also row scaling
    #
    cdef double acc    # accumulator for summation
    cdef int j, oj, k  # loop counters
    if do_sens:
        for j in range(nr):      # row in reduced system
            oj = r2o[j]  # index of reduced DOF j in original system

            acc = 0.
            for k in range(nk):
                acc += w[k] * fk[k] * c[ k*no + oj ]  # apply also w[k], the fitting weight of error component k

                # sensitivity of the solution w.r.t. fk[k]
                # (only one term in each row; this is d(b) / d( fk[k] ), leaving only c^(j)_k)
                s[j + nr*k] = row_scale[j] * w[k] * c[ k*no + oj ]
            b[j] = row_scale[j] * acc
    else:
        for j in range(nr):      # row in reduced system
            oj = r2o[j]  # index of reduced DOF j in original system

            acc = 0.
            for k in range(nk):
                acc += w[k] * fk[k] * c[ k*no + oj ]   # apply also w[k], the fitting weight of error component k
            b[j] = row_scale[j] * acc

    # apply algebraic elimination to the knowns
    #
    cdef int om
    for om in range(no):  # DOF in original system
        # handle only the eliminated DOFs
        if knowns & (1LL << om):
            # om now represents an eliminated column of the original system

            # update all rows of b in the reduced system.
            #
            # We must account for row scaling, because the whole row is scaled, and hence b, too.
            #
            # But column scaling cancels out for eliminated DOFs:
            #
            #   fi[om] * c[ k*SIZE + om ] = (fi[om] * col_scale) * (c[ k*SIZE + om ] / col_scale)
            #   ======   ================   --------------------   ------------------------------
            #   "small"    original coeff   "big" (scaled) units   scaled (big) coeff in A
            #   (orig.)
            #    units
            #
            # or intuitively:
            #
            #   many small units * small coeff = few big units * scaled (big) coeff
            #
            # This is fortunate, because the column scaling factor has not been computed for the eliminated columns!
            #
            for j in range(nr):  # row in reduced system
                oj = r2o[j]  # index of reduced DOF j in original system
                for k in range(nk):
                    b[j] -= fi[om] * w[k] * c[ k*no + om ] * c[ k*no + oj ] * row_scale[j]   # apply also w[k], the fitting weight of error component k

            # also, we may as well do this here:
            if do_sens:  # no is rather small, not many extra instructions
                for k in range(nk):
                    sens[k, om] = nan  # no sensitivity data for eliminated DOFs

    # solve (overwriting b with solution)
    #
    drivers.generalfactored_c( LU, ipiv, b, nr )

    # perform sensitivity analysis (overwriting s with the sensitivity data)
    #
    if do_sens:  # if no sensitivity analysis, major runtime savings occur here
        # we must loop as there is no "generalfactoreds_c()".
        for k in range(nk):
            drivers.generalfactored_c( LU, ipiv, &s[k*nr], nr )

    # extract solution (filling in only the unknowns in fi[])
    #
    for j in range(nr):  # row in reduced system
        oj = r2o[j]  # index of reduced DOF j in original system

        fi[oj] = b[j] * col_scale[j]  # compute original x from the column-scaled x

        # fill sensitivity data w.r.t. each unknown j and each fk[k]
        if do_sens:  # again, small nr, not many extra instructions
            for k in range(nk):
                sens[k, oj] = s[nr*k + j] * col_scale[j]


# Algorithm with iterative refinement of the fit to mitigate roundoff.
#
# Uses solve() to do the actual solving.
#
# Note that the xk data (the same neighbor points as used in the fitting) must be passed in here, so that the fitted model can be interpolated to the correct points when computing the error.
# The model origin xi is stored in the Case object, but it would make no sense to store the xk data in the Case, because this data is in practice likely shared between problem instances.
#
# In any case, the data stored in the Case object is dependent on xk and xi not changing on the fly. If xi or xk suddenly change, the c and A matrices must be re-generated.
#
cdef int solve_iterative( infra.Case* case, double[::view.generic] fk, double[::view.generic,::view.contiguous] sens, int do_sens, int taskid, int max_iter,
                          double[::view.generic,::view.contiguous] xkManyD, double[::view.generic] xk1D ) nogil:

    DEF DONT_DO_SENS = 0  # value of do_sens when we don't want to do sensitivity analysis (used in the iterative refinement step)
    DEF DIFF = 0  # interpolate function value (not a derivative)

    # First, as usual, fit the model against the original user-given data fk[], automatically updating case.fi:
    #
    solve( case, fk, sens, do_sens, taskid )

    # Then refine iteratively to reduce numerical error.

    # Get temporary arrays for work space, to be used for fitting the remaining error.
    cdef double* wrk_fk = infra.Case_get_fk_tmp( case, taskid )
    cdef double* wrk_fi = infra.Case_get_fi_tmp( case, taskid )
    cdef int nk = case.nk, no = case.no

    # Zero out those coefficients of the "error reduction" fit for which the solution is known exactly,
    # since we cannot change those to reduce the error. (They will be tagged as knowns for the
    # error reduction problem, too.)
    #
    cdef long long knowns = case.knowns
    cdef int om
    for om in range(no):
        if knowns & (1LL << om):
            wrk_fi[om] = 0.  # this component of the solution has no error to reduce

    cdef int i=0, k  # the user may have accidentally given max_iter=0 so we must initialize i here
    cdef double tmp, norm, prev_norm=-1.
    cdef int dimension = case.dimension
    cdef int order     = case.order
    cdef double* fi    = case.fi  # the array has case.no elements
#    DEF epsilon = 1e-15
    for i in range(max_iter):
        # Using the latest fi[], interpolate the fitted model to the points xk, overwriting wrk_fk.
        #
        # This sets up the new target for fitting (i.e. remaining error at the points xk).
        #
        interp.interpolate_nD( case, xkManyD, xk1D, wrk_fk, DIFF )  # only either *ManyD or *1D are used, depending on case.dimension
        for k in range(nk):
            wrk_fk[k] = fk[k] - wrk_fk[k]  # overwrite wrk_fk a second time, now with the remaining error


        # Convergence check: monitor the l-infinity norm (max abs) of remaining error
        #
        # Note that if we get very lucky (and the data we are fitting against allows it),
        # at i=0, the error may already be below epsilon.
        #
        # In such a case we break immediately before performing any error reduction fit;
        # thus the optimal ordering is to check convergence *before* actually fitting.
        #
        # TODO: What is a good error measure? |f| may be very large, or it may be near (or at!) zero...
        # TODO: Or is it better to always run, say, 3 iterations and be done with it? (would simplify the code)
        #
        norm = c_abs(wrk_fk[0])
        for k in range(1,nk):
            tmp = c_abs(wrk_fk[k])
            if tmp > norm:
                norm = tmp

#        # This is useless as it depends on the scaling of the problem (size of function values fk).
#        #
#        # Checking that the l-infinity norm of the error no longer changes should be good enough also in cases where the iteration converges to machine accuracy.
#        #
#        # If we really want to do this, we should probably normalize by the l-infinity norm of fk. But that runs into trouble if all fk are very small.
#        if norm <= epsilon:  # converged, error successfully eliminated
#            break

        # If the norm is the same as in the previous iteration (down to floating point equality!),
        # we have probably stagnated (i.e. the remaining error is no longer changing).
        #
        # In this case iterative refinement does not reduce the error any further (the remaining error is likely to be modeling error, not roundoff!),
        # and we can consider the fitting complete.
        #
        if norm == prev_norm:  # converged, further corrections do nothing
                               # (when we get here, we have done the error reduction fit at least once and the norm didn't change)
                               # (the initial prev_norm is an invalid value so this cannot trigger at the first iteration)
            break
        prev_norm = norm


        # Fit the coefficients wrk_fi to model the remaining error wrk_fk,
        # using the same type of model as used for the original fit.
        #
        # To avoid the need for GIL, use the solver version that accepts raw pointers for fk and fi.
        #
        # Note that for the coefficient array, we use wrk_fi, not case.fi! This is important to keep the original solution intact.
        #
        solve_contig( case, wrk_fk, wrk_fi, sens, DONT_DO_SENS, taskid )

        # Update the unknown components of the solution fi
        # to make the model describe  original data + error .
        #
        for om in range(no):
            if not (knowns & (1LL << om)):
                fi[om] += wrk_fi[om]

    else:  # if the loop didn't break, the last iteration of the loop contributed one more refinement iteration
        i += 1

    return i  # number of refinement iterations taken (i, not i+1, because the fitting code comes *after* the convergence check)

